import io
import os
from typing import List, Optional, Tuple
from unittest import mock
from zipfile import ZipFile

import pytest
from pydantic import ValidationError, root_validator

from cumulusci.core.config.org_config import OrgConfig, VersionInfo
from cumulusci.core.config.project_config import BaseProjectConfig
from cumulusci.core.dependencies.base import (
    DynamicDependency,
    StaticDependency,
    UnmanagedVcsDependencyFlow,
)
from cumulusci.core.dependencies.dependencies import (
    PackageNamespaceVersionDependency,
    PackageVersionIdDependency,
    UnmanagedZipURLDependency,
    parse_dependency,
)
from cumulusci.core.dependencies.github import (
    GitHubDynamicDependency,
    GitHubDynamicSubfolderDependency,
    UnmanagedGitHubRefDependency,
)
from cumulusci.core.dependencies.resolvers import (
    AbstractResolver,
    DependencyResolutionStrategy,
)
from cumulusci.core.exceptions import DependencyResolutionError
from cumulusci.salesforce_api.package_install import (
    DEFAULT_PACKAGE_RETRY_OPTIONS,
    PackageInstallOptions,
)
from cumulusci.utils.version_strings import StrictVersion
from cumulusci.utils.ziputils import zip_subfolder


def _sync_vcs_and_url(values):
    # If only vcs is provided, set url to vcs
    if values.get("vcs") and not values.get("url"):
        values["url"] = values["vcs"]
    # If only url is provided, set vcs to url
    elif values.get("url") and not values.get("vcs"):
        values["vcs"] = values["url"]
    return values


class ConcreteDynamicDependency(DynamicDependency):
    ref: Optional[str]
    resolved: Optional[bool] = False
    vcs: str = "github"

    @property
    def is_resolved(self):
        return self.resolved

    def resolve(
        self, context: BaseProjectConfig, strategies: List[DependencyResolutionStrategy]
    ):
        super().resolve(context, strategies)
        self.resolved = True

    @property
    def name(self):
        return ""

    @root_validator(pre=True)
    def sync_vcs_and_url(cls, values):
        """Defined vcs should be assigned to url"""
        return _sync_vcs_and_url(values)


class MockResolver(AbstractResolver):
    def __init__(
        self,
        resolve_ref: Optional[str] = None,
        resolve_dep: Optional[StaticDependency] = None,
    ):
        self.ref = resolve_ref
        self.dep = resolve_dep

    def can_resolve(self, dep: DynamicDependency, context: BaseProjectConfig) -> bool:
        return bool(self.dep and self.ref)

    def resolve(
        self, dep: DynamicDependency, context: BaseProjectConfig
    ) -> Tuple[Optional[str], Optional[StaticDependency]]:
        return self.ref, self.dep


class MockBadResolver(AbstractResolver):
    def can_resolve(self, dep: DynamicDependency, context: BaseProjectConfig) -> bool:
        return True

    def resolve(
        self, dep: DynamicDependency, context: BaseProjectConfig
    ) -> Tuple[Optional[str], Optional[StaticDependency]]:
        raise DependencyResolutionError("Bad resolver")


class TestDynamicDependency:
    @mock.patch("cumulusci.core.dependencies.resolvers.get_resolver")
    def test_dynamic_dependency(self, get_resolver):
        d = ConcreteDynamicDependency()
        resolvers = [
            MockResolver(),
            MockResolver(
                "aaaaaaaaaaaaaaaa",
                PackageNamespaceVersionDependency(namespace="foo", version="1.0"),
            ),
        ]
        get_resolver.side_effect = resolvers

        d.resolve(
            mock.Mock(),
            [
                DependencyResolutionStrategy.UNMANAGED_HEAD,
                DependencyResolutionStrategy.COMMIT_STATUS_PREVIOUS_RELEASE_BRANCH,
            ],
        )

        assert d.package_dependency == PackageNamespaceVersionDependency(
            namespace="foo", version="1.0"
        )
        assert d.ref == "aaaaaaaaaaaaaaaa"

    @mock.patch("cumulusci.core.dependencies.resolvers.get_resolver")
    def test_dynamic_dependency__twice(self, get_resolver):
        d = ConcreteDynamicDependency()
        resolvers = [
            mock.Mock(
                wraps=MockResolver(
                    "aaaaaaaaaaaaaaaa",
                    PackageNamespaceVersionDependency(namespace="foo", version="1.0"),
                )
            ),
        ]
        get_resolver.side_effect = resolvers

        d.resolve(
            mock.Mock(),
            [
                DependencyResolutionStrategy.UNMANAGED_HEAD,
                DependencyResolutionStrategy.COMMIT_STATUS_PREVIOUS_RELEASE_BRANCH,
            ],
        )

        assert d.is_resolved
        resolvers[0].resolve.assert_called_once()
        d.resolve(
            mock.Mock(),
            [
                DependencyResolutionStrategy.UNMANAGED_HEAD,
                DependencyResolutionStrategy.COMMIT_STATUS_PREVIOUS_RELEASE_BRANCH,
            ],
        )
        assert d.is_resolved
        resolvers[0].resolve.assert_called_once()

    @mock.patch("cumulusci.core.dependencies.resolvers.get_resolver")
    def test_dynamic_dependency_resolution_fails(self, get_resolver):
        d = ConcreteDynamicDependency()
        resolvers = [MockBadResolver()]
        get_resolver.side_effect = resolvers

        with pytest.raises(DependencyResolutionError):
            d.resolve(mock.Mock(), [DependencyResolutionStrategy.UNMANAGED_HEAD])

    @mock.patch("cumulusci.core.dependencies.resolvers.get_resolver")
    def test_dynamic_dependency_resolution_no_results(self, get_resolver):
        d = ConcreteDynamicDependency()
        resolvers = [MockResolver("", None)]
        get_resolver.side_effect = resolvers

        with pytest.raises(DependencyResolutionError):
            d.resolve(mock.Mock(), [DependencyResolutionStrategy.UNMANAGED_HEAD])


class TestGitHubDynamicSubfolderDependency:
    def test_flatten(self):
        gh = GitHubDynamicSubfolderDependency(
            github="https://github.com/Test/TestRepo", subfolder="foo"
        )

        gh.ref = "aaaa"

        assert gh.is_unmanaged
        assert gh.flatten(mock.Mock()) == [
            UnmanagedGitHubRefDependency(
                github="https://github.com/Test/TestRepo",
                subfolder="foo",
                ref="aaaa",
                namespace_inject=None,
                namespace_strip=None,
            )
        ]

    def test_flatten__unresolved(self):
        context = mock.Mock()
        gh = GitHubDynamicSubfolderDependency(
            repo_owner="Test", repo_name="TestRepo", subfolder="foo"
        )

        with pytest.raises(DependencyResolutionError) as e:
            gh.flatten(context)

        assert "is not resolved" in str(e)

    def test_name(self):
        gh = GitHubDynamicSubfolderDependency(
            repo_owner="Test", repo_name="TestRepo", subfolder="foo"
        )
        assert gh.github in gh.name and gh.subfolder in gh.name

    def test_description(self):
        gh = GitHubDynamicSubfolderDependency(
            repo_owner="Test", repo_name="TestRepo", subfolder="foo"
        )
        gh.ref = "aaaa"
        assert (
            gh.github in gh.description
            and gh.subfolder in gh.description
            and gh.ref in gh.description
        )


class TestGitHubDynamicDependency:
    def test_create_repo_name(self):
        gh = GitHubDynamicDependency(repo_owner="Test", repo_name="TestRepo")
        assert gh.github == "https://github.com/Test/TestRepo"

    def test_create_failure(self):
        with pytest.raises(ValidationError):
            GitHubDynamicDependency(repo_owner="Test")

        with pytest.raises(ValidationError):
            GitHubDynamicDependency(
                github="http://github.com/Test/TestRepo", tag="tag/1.0", ref="aaaaa"
            )

        with pytest.raises(ValidationError):
            GitHubDynamicDependency(
                github="http://github.com/Test/TestRepo", namespace_inject="foo"
            )

    @mock.patch("cumulusci.core.dependencies.github.get_github_repo")
    def test_flatten(self, repo, project_config):
        repo.side_effect = project_config.get_github_repo_side_effect

        gh = GitHubDynamicDependency(github="https://github.com/SFDO-Tooling/RootRepo")
        gh.ref = "aaaaa"
        gh.package_dependency = PackageNamespaceVersionDependency(
            namespace="bar", version="2.0"
        )

        root_repo = PackageNamespaceVersionDependency(namespace="bar", version="2.0")

        assert gh.flatten(project_config) == [
            GitHubDynamicDependency(
                github="https://github.com/SFDO-Tooling/DependencyRepo",
                password_env_name="DEP_PW",
            ),
            UnmanagedGitHubRefDependency(
                github="https://github.com/SFDO-Tooling/RootRepo",
                subfolder="unpackaged/pre/first",
                unmanaged=True,
                ref="aaaaa",
                package_dependency=root_repo,
            ),
            UnmanagedGitHubRefDependency(
                github="https://github.com/SFDO-Tooling/RootRepo",
                subfolder="unpackaged/pre/second",
                unmanaged=True,
                ref="aaaaa",
                package_dependency=root_repo,
            ),
            root_repo,
            UnmanagedGitHubRefDependency(
                github="https://github.com/SFDO-Tooling/RootRepo",
                subfolder="unpackaged/post/first",
                unmanaged=False,
                ref="aaaaa",
                namespace_inject="bar",
                package_dependency=root_repo,
            ),
        ]

    @mock.patch("cumulusci.core.dependencies.github.get_github_repo")
    def test_flatten__skip(self, repo, project_config):
        repo.side_effect = project_config.get_github_repo_side_effect
        gh = GitHubDynamicDependency(
            github="https://github.com/SFDO-Tooling/RootRepo",
            skip="unpackaged/pre/first",
        )
        gh.ref = "aaaaa"
        gh.package_dependency = PackageNamespaceVersionDependency(
            namespace="bar", version="2.0"
        )

        root_repo = PackageNamespaceVersionDependency(namespace="bar", version="2.0")

        assert gh.flatten(project_config) == [
            GitHubDynamicDependency(
                github="https://github.com/SFDO-Tooling/DependencyRepo",
                password_env_name="DEP_PW",
            ),
            UnmanagedGitHubRefDependency(
                github="https://github.com/SFDO-Tooling/RootRepo",
                subfolder="unpackaged/pre/second",
                unmanaged=True,
                ref="aaaaa",
                package_dependency=root_repo,
            ),
            root_repo,
            UnmanagedGitHubRefDependency(
                github="https://github.com/SFDO-Tooling/RootRepo",
                subfolder="unpackaged/post/first",
                unmanaged=False,
                ref="aaaaa",
                namespace_inject="bar",
                package_dependency=root_repo,
            ),
        ]

    @mock.patch("cumulusci.core.dependencies.github.get_github_repo")
    def test_flatten__not_found(self, repo, project_config):
        repo.side_effect = project_config.get_github_repo_side_effect
        gh = GitHubDynamicDependency(
            github="https://github.com/SFDO-Tooling/NoUnmanagedPreRepo",
        )
        gh.ref = "aaaaa"
        gh.package_dependency = PackageNamespaceVersionDependency(
            namespace="foo", version="2.0"
        )

        assert gh.flatten(project_config) == [
            PackageNamespaceVersionDependency(namespace="foo", version="2.0"),
        ]

    def test_flatten__unresolved(self):
        context = mock.Mock()
        gh = GitHubDynamicDependency(repo_owner="Test", repo_name="TestRepo")

        with pytest.raises(DependencyResolutionError) as e:
            gh.flatten(context)

        assert "is not resolved" in str(e)

    @mock.patch("cumulusci.core.dependencies.github.get_github_repo")
    def test_flatten__bad_transitive_dep(self, repo, project_config):
        repo.side_effect = project_config.get_github_repo_side_effect
        gh = GitHubDynamicDependency(repo_owner="Test", repo_name="RootRepoBadDep")
        gh.ref = "aaaa"

        with pytest.raises(DependencyResolutionError) as e:
            gh.flatten(project_config)

        assert "transitive dependency could not be parsed" in str(e)

    @mock.patch("cumulusci.core.dependencies.github.get_github_repo")
    def test_flatten__unmanaged_src(self, repo, project_config):
        repo.side_effect = project_config.get_github_repo_side_effect
        gh = GitHubDynamicDependency(
            github="https://github.com/SFDO-Tooling/RootRepo",
            unmanaged=True,
        )
        gh.ref = "aaaaa"

        assert gh.flatten(project_config) == [
            GitHubDynamicDependency(
                github="https://github.com/SFDO-Tooling/DependencyRepo",
                password_env_name="DEP_PW",
            ),
            UnmanagedGitHubRefDependency(
                github="https://github.com/SFDO-Tooling/RootRepo",
                subfolder="unpackaged/pre/first",
                unmanaged=True,
                ref="aaaaa",
            ),
            UnmanagedGitHubRefDependency(
                github="https://github.com/SFDO-Tooling/RootRepo",
                subfolder="unpackaged/pre/second",
                unmanaged=True,
                ref="aaaaa",
            ),
            UnmanagedGitHubRefDependency(
                github="https://github.com/SFDO-Tooling/RootRepo",
                unmanaged=True,
                ref="aaaaa",
            ),
            UnmanagedGitHubRefDependency(
                github="https://github.com/SFDO-Tooling/RootRepo",
                subfolder="unpackaged/post/first",
                unmanaged=True,
                ref="aaaaa",
                namespace_strip="bar",
            ),
        ]

    @mock.patch("cumulusci.core.dependencies.github.get_github_repo")
    def test_flatten__no_release(self, repo, project_config):
        repo.side_effect = project_config.get_github_repo_side_effect
        gh = GitHubDynamicDependency(
            github="https://github.com/SFDO-Tooling/RootRepo",
            unmanaged=False,
        )
        gh.ref = "aaaaa"

        with pytest.raises(DependencyResolutionError) as e:
            gh.flatten(project_config)

        assert "Could not find latest release" in str(e)

    def test_name(self):
        gh = GitHubDynamicDependency(github="https://github.com/SFDO-Tooling/RootRepo")
        assert gh.github in gh.name


class TestPackageNamespaceVersionDependency:
    @mock.patch(
        "cumulusci.core.dependencies.dependencies.install_package_by_namespace_version"
    )
    def test_install(self, install_package_by_namespace_version):
        m = PackageNamespaceVersionDependency(namespace="test", version="1.0")

        context = mock.Mock()
        org = OrgConfig({}, "dev")
        org._installed_packages = {}

        m.install(context, org)

        install_package_by_namespace_version.assert_called_once_with(
            context,
            org,
            m.namespace,
            m.version,
            PackageInstallOptions(),
            retry_options=DEFAULT_PACKAGE_RETRY_OPTIONS,
        )

    @mock.patch(
        "cumulusci.core.dependencies.dependencies.install_package_by_namespace_version"
    )
    def test_install__already_installed(self, install_package_by_namespace_version):
        m = PackageNamespaceVersionDependency(namespace="test", version="1.0")

        context = mock.Mock()
        org = OrgConfig({}, "dev")
        org._installed_packages = {
            "test": [VersionInfo(id="04t000000000000", number=StrictVersion("1.0"))]
        }

        m.install(context, org)

        install_package_by_namespace_version.assert_not_called()

    @mock.patch(
        "cumulusci.core.dependencies.dependencies.install_package_by_namespace_version"
    )
    def test_install__newer_beta(self, install_package_by_namespace_version):
        m = PackageNamespaceVersionDependency(namespace="test", version="1.1 (Beta 4)")

        context = mock.Mock()
        org = OrgConfig({}, "dev")
        org._installed_packages = {
            "test": [VersionInfo(id="04t000000000000", number=StrictVersion("1.0"))]
        }

        m.install(context, org)

        install_package_by_namespace_version.assert_called_once_with(
            context,
            org,
            m.namespace,
            m.version,
            PackageInstallOptions(),
            retry_options=DEFAULT_PACKAGE_RETRY_OPTIONS,
        )

    @mock.patch(
        "cumulusci.core.dependencies.dependencies.install_package_by_namespace_version"
    )
    def test_install__custom_options(self, install_package_by_namespace_version):
        m = PackageNamespaceVersionDependency(namespace="foo", version="1.0")

        context = mock.Mock()
        org = OrgConfig({}, "dev")
        org._installed_packages = {}
        opts = PackageInstallOptions(password="test")

        m.install(context, org, options=opts)

        install_package_by_namespace_version.assert_called_once_with(
            context,
            org,
            m.namespace,
            m.version,
            opts,
            retry_options=DEFAULT_PACKAGE_RETRY_OPTIONS,
        )

    @mock.patch(
        "cumulusci.core.dependencies.dependencies.install_package_by_namespace_version"
    )
    def test_install__key_from_env(self, install_package_by_namespace_version):
        m = PackageNamespaceVersionDependency(
            namespace="foo", version="1.0", password_env_name="PW"
        )

        context = mock.Mock()
        org = OrgConfig({}, "dev")
        org._installed_packages = {}

        with mock.patch.dict(os.environ, PW="testpw"):
            m.install(context, org)

        opts = install_package_by_namespace_version.call_args[0][4]
        assert opts.password == "testpw"

    def test_name(self):
        assert (
            PackageNamespaceVersionDependency(namespace="foo", version="1.0").name
            == "Install foo 1.0"
        )

    def test_package_name(self):
        assert (
            PackageNamespaceVersionDependency(
                namespace="foo", version="1.0", package_name="Foo"
            ).package
            == "Foo"
        )

        assert (
            PackageNamespaceVersionDependency(namespace="foo", version="1.0").package
            == "foo"
        )


class TestPackageVersionIdDependency:
    @mock.patch(
        "cumulusci.core.dependencies.dependencies.install_package_by_version_id"
    )
    def test_install(self, install_package_by_version_id):
        m = PackageVersionIdDependency(version_id="04t000000000000")

        context = mock.Mock()
        org = OrgConfig({}, "dev")
        org._installed_packages = {}

        m.install(context, org)

        install_package_by_version_id.assert_called_once_with(
            context,
            org,
            m.version_id,
            PackageInstallOptions(),
            retry_options=DEFAULT_PACKAGE_RETRY_OPTIONS,
        )

    @mock.patch(
        "cumulusci.core.dependencies.dependencies.install_package_by_version_id"
    )
    def test_install__already_installed(self, install_package_by_version_id):
        m = PackageVersionIdDependency(version_id="04t000000000000")

        context = mock.Mock()
        org = OrgConfig({}, "dev")
        org._installed_packages = {
            "04t000000000000": [VersionInfo(number="1.0", id="04t000000000000")]
        }

        m.install(context, org)

        install_package_by_version_id.assert_not_called()

    @mock.patch(
        "cumulusci.core.dependencies.dependencies.install_package_by_version_id"
    )
    def test_install__custom_options(self, install_package_by_version_id):
        m = PackageVersionIdDependency(version_id="04t000000000000")

        context = mock.Mock()
        org = OrgConfig({}, "dev")
        org._installed_packages = {}
        opts = PackageInstallOptions(password="test")

        m.install(context, org, options=opts)

        install_package_by_version_id.assert_called_once_with(
            context,
            org,
            m.version_id,
            opts,
            retry_options=DEFAULT_PACKAGE_RETRY_OPTIONS,
        )

    @mock.patch(
        "cumulusci.core.dependencies.dependencies.install_package_by_version_id"
    )
    def test_install__key_from_env(self, install_package_by_version_id):
        m = PackageVersionIdDependency(
            version_id="04t000000000000", password_env_name="PW"
        )

        context = mock.Mock()
        org = OrgConfig({}, "dev")
        org._installed_packages = {}

        with mock.patch.dict(os.environ, PW="testpw"):
            m.install(context, org)

        opts = install_package_by_version_id.call_args[0][3]
        assert opts.password == "testpw"

    def test_name(self):
        assert (
            PackageVersionIdDependency(
                package_name="foo", version_id="04t000000000000"
            ).name
            == "Install foo 04t000000000000"
        )

    def test_package_name(self):
        assert (
            PackageVersionIdDependency(version_id="04t000000000000").package
            == "Unknown Package"
        )


class TestUnmanagedGitHubRefDependency:
    def test_validation(self):
        with pytest.raises(ValidationError):
            UnmanagedGitHubRefDependency(github="http://github.com")

        u = UnmanagedGitHubRefDependency(
            github="https://github.com/Test/TestRepo",
            ref="aaaaaaaa",
            namespace_token="obsolete but accepted",
            filename_token="obsolete but accepted",
        )

        u = UnmanagedGitHubRefDependency(
            repo_owner="Test", repo_name="TestRepo", ref="aaaaaaaa"
        )
        assert u.github == "https://github.com/Test/TestRepo"

    @mock.patch("cumulusci.core.dependencies.github.get_github_repo")
    @mock.patch("cumulusci.core.dependencies.base.download_extract_vcs_from_repo")
    @mock.patch("cumulusci.core.dependencies.base.MetadataPackageZipBuilder")
    @mock.patch("cumulusci.core.dependencies.base.ApiDeploy")
    def test_install(
        self, api_deploy_mock, zip_builder_mock, download_mock, repo_mock, init_git_repo
    ):
        repo_mock.return_value = init_git_repo
        d = UnmanagedGitHubRefDependency(
            github="http://github.com/Test/TestRepo", ref="aaaaaaaa"
        )

        with ZipFile(io.BytesIO(), "w") as zf:
            zf.writestr("package.xml", "test")
            download_mock.return_value = zf

            context = mock.Mock()
            org = mock.Mock()
            d.install(context, org)

            download_mock.assert_called_once_with(init_git_repo, ref=d.ref)
            zip_builder_mock.from_zipfile.assert_called_once_with(
                download_mock.return_value,
                path=None,
                options={
                    "unmanaged": True,
                    "namespace_inject": None,
                    "namespace_strip": None,
                },
                context=mock.ANY,
            )
            api_deploy_mock.assert_called_once_with(
                mock.ANY,  # The context object is checked below
                zip_builder_mock.from_zipfile.return_value.as_base64.return_value,
            )
            mock_task = api_deploy_mock.call_args_list[0][0][0]
            assert mock_task.org_config == org
            assert mock_task.project_config == context

            api_deploy_mock.return_value.assert_called_once()
            zf.close()

    def test_get_unmanaged(self):
        org = mock.Mock()
        org.installed_packages = {"foo": "1.0"}
        assert (
            UnmanagedGitHubRefDependency(
                github="http://github.com/Test/TestRepo", ref="aaaa", unmanaged=True
            )._get_unmanaged(org)
            is True
        )
        assert (
            UnmanagedGitHubRefDependency(
                github="http://github.com/Test/TestRepo", ref="aaaa"
            )._get_unmanaged(org)
            is True
        )
        assert (
            UnmanagedGitHubRefDependency(
                github="http://github.com/Test/TestRepo",
                ref="aaaa",
                namespace_inject="foo",
            )._get_unmanaged(org)
            is False
        )
        assert (
            UnmanagedGitHubRefDependency(
                github="http://github.com/Test/TestRepo",
                ref="aaaa",
                namespace_inject="bar",
            )._get_unmanaged(org)
            is True
        )

    def test_name(self):
        assert (
            UnmanagedGitHubRefDependency(
                github="http://github.com/Test/TestRepo",
                subfolder="unpackaged/pre/first",
                ref="aaaa",
            ).name
            == "Deploy http://github.com/Test/TestRepo/unpackaged/pre/first"
        )

        assert (
            UnmanagedGitHubRefDependency(
                github="http://github.com/Test/TestRepo",
                ref="aaaa",
            ).name
            == "Deploy http://github.com/Test/TestRepo"
        )


class TestUnmanagedZipURLDependency:
    @mock.patch("cumulusci.core.dependencies.dependencies.download_extract_zip")
    @mock.patch("cumulusci.core.dependencies.base.MetadataPackageZipBuilder")
    @mock.patch("cumulusci.core.dependencies.base.ApiDeploy")
    def test_install(self, api_deploy_mock, zip_builder_mock, download_mock):
        d = UnmanagedZipURLDependency(zip_url="http://foo.com")

        with ZipFile(io.BytesIO(), "w") as zf:
            zf.writestr("src/package.xml", "test")
            download_mock.return_value = zf

            context = mock.Mock()
            org = mock.Mock()
            d.install(context, org)

            download_mock.assert_called_once_with(d.zip_url)

            zip_builder_mock.from_zipfile.assert_called_once_with(
                mock.ANY,
                options={
                    "unmanaged": True,
                    "namespace_inject": None,
                    "namespace_strip": None,
                },
                path=None,
                context=mock.ANY,
            )
            api_deploy_mock.assert_called_once_with(
                mock.ANY,  # The context object is checked below
                zip_builder_mock.from_zipfile.return_value.as_base64.return_value,
            )
            mock_task = api_deploy_mock.call_args_list[0][0][0]
            assert mock_task.org_config == org
            assert mock_task.project_config == context

            api_deploy_mock.return_value.assert_called_once()
            zf.close()

    def test_get_unmanaged(self):
        org = mock.Mock()
        org.installed_packages = {"foo": "1.0"}
        assert (
            UnmanagedZipURLDependency(
                zip_url="http://foo.com", unmanaged=True
            )._get_unmanaged(org)
            is True
        )
        assert (
            UnmanagedZipURLDependency(
                zip_url="http://foo.com", namespace_inject="foo"
            )._get_unmanaged(org)
            is False
        )
        assert (
            UnmanagedZipURLDependency(
                zip_url="http://foo.com", namespace_inject="bar"
            )._get_unmanaged(org)
            is True
        )

    def test_name(self):
        assert (
            UnmanagedZipURLDependency(zip_url="http://foo.com", subfolder="bar").name
            == "Deploy http://foo.com /bar"
        )

    @mock.patch("cumulusci.core.dependencies.base.MetadataPackageZipBuilder")
    @mock.patch("cumulusci.core.dependencies.dependencies.download_extract_zip")
    @mock.patch("cumulusci.core.dependencies.base.zip_subfolder")
    def test_get_metadata_package_zip_builder__mdapi_root(
        self, subfolder_mock, download_zip_mock, zipbuilder_mock
    ):
        with ZipFile(io.BytesIO(), "w") as zf:
            zf.writestr("src/package.xml", "test")

            dep = UnmanagedZipURLDependency(zip_url="http://foo.com")
            download_zip_mock.return_value = zf
            subfolder_mock.return_value = zip_subfolder(zf, "src")

            context = mock.Mock()
            org = mock.Mock()

            assert (
                dep.get_metadata_package_zip_builder(context, org)
                == zipbuilder_mock.from_zipfile.return_value
            )
            subfolder_mock.assert_called_once_with(zf, "src")
            zipbuilder_mock.from_zipfile.assert_called_once_with(
                subfolder_mock.return_value,
                path=None,
                options={
                    "unmanaged": True,
                    "namespace_inject": None,
                    "namespace_strip": None,
                },
                context=mock.ANY,
            )
            zf.close()

    @mock.patch("cumulusci.core.dependencies.base.MetadataPackageZipBuilder")
    @mock.patch("cumulusci.core.dependencies.dependencies.download_extract_zip")
    @mock.patch("cumulusci.core.dependencies.base.zip_subfolder")
    def test_get_metadata_package_zip_builder__mdapi_subfolder(
        self, subfolder_mock, download_zip_mock, zipbuilder_mock
    ):
        with ZipFile(io.BytesIO(), "w") as zf:

            zf.writestr("unpackaged/pre/first/package.xml", "test")

            dep = UnmanagedZipURLDependency(
                zip_url="http://foo.com", subfolder="unpackaged/pre/first"
            )
            download_zip_mock.return_value = zf
            subfolder_mock.return_value = zip_subfolder(zf, "unpackaged/pre/first")
            context = mock.Mock()
            org = mock.Mock()

            assert (
                dep.get_metadata_package_zip_builder(context, org)
                == zipbuilder_mock.from_zipfile.return_value
            )
            subfolder_mock.assert_called_once_with(zf, "unpackaged/pre/first")
            zipbuilder_mock.from_zipfile.assert_called_once_with(
                subfolder_mock.return_value,
                path=None,
                options={
                    "unmanaged": True,
                    "namespace_inject": None,
                    "namespace_strip": None,
                },
                context=mock.ANY,
            )
            zf.close()

    @mock.patch("cumulusci.core.dependencies.base.MetadataPackageZipBuilder")
    @mock.patch("cumulusci.core.dependencies.dependencies.download_extract_zip")
    @mock.patch("cumulusci.core.dependencies.base.zip_subfolder")
    @mock.patch("cumulusci.core.sfdx.sfdx")
    def test_get_metadata_package_zip_builder__sfdx(
        self, sfdx_mock, subfolder_mock, download_zip_mock, zipbuilder_mock
    ):
        zf = ZipFile(io.BytesIO(), "w")

        zf.writestr("force-app/main/default/classes/", "test")

        dep = UnmanagedZipURLDependency(zip_url="http://foo.com", subfolder="force-app")
        download_zip_mock.return_value = zf

        context = mock.Mock()
        org = mock.Mock()

        assert (
            dep.get_metadata_package_zip_builder(context, org)
            == zipbuilder_mock.from_zipfile.return_value
        )
        subfolder_mock.assert_not_called()
        zipbuilder_mock.from_zipfile.assert_called_once_with(
            None,
            path=mock.ANY,
            options={
                "unmanaged": True,
                "namespace_inject": None,
                "namespace_strip": None,
            },
            context=mock.ANY,
        )
        sfdx_mock.assert_called_once_with(
            "project convert source",
            args=["-d", mock.ANY, "-r", "force-app"],
            capture_output=True,
            check_return=True,
        )
        zf.close()


# Concrete implementation of UnmanagedVcsDependencyFlow for testing
class ConcreteUnmanagedVcsDependencyFlow(UnmanagedVcsDependencyFlow):
    """Concrete implementation of UnmanagedVcsDependencyFlow for testing"""

    def __init__(self, **kwargs):
        # Set default values for testing
        defaults = {
            "vcs": "github",
            "url": "https://github.com/test/repo",
            "commit": "abc123",
            "flow_name": "test_flow",
        }
        defaults.update(kwargs)
        super().__init__(**defaults)

    @root_validator(pre=True)
    def sync_vcs_and_url(cls, values):
        """Implement required abstract method from base class"""
        return _sync_vcs_and_url(values)


class TestUnmanagedVcsDependencyFlow:
    def test_name(self):
        """Test the name property"""
        flow_dep = ConcreteUnmanagedVcsDependencyFlow(
            url="https://github.com/test/repo", flow_name="install_deps"
        )
        expected_name = "Deploy https://github.com/test/repo Flow: install_deps"
        assert flow_dep.name == expected_name

    def test_description(self):
        """Test the description property"""
        flow_dep = ConcreteUnmanagedVcsDependencyFlow(
            url="https://github.com/test/repo",
            flow_name="install_deps",
            commit="abc123",
        )
        expected_description = "https://github.com/test/repo Flow: install_deps @abc123"
        assert flow_dep.description == expected_description

    def test_get_unmanaged(self):
        """Test the _get_unmanaged method inherited from UnmanagedStaticDependency"""
        # Test with unmanaged = None and no namespace_inject
        flow_dep = ConcreteUnmanagedVcsDependencyFlow(
            unmanaged=None, namespace_inject=None
        )
        org_config = mock.Mock()
        org_config.installed_packages = {}
        assert flow_dep._get_unmanaged(org_config) is True

        # Test with unmanaged = None and namespace_inject present but not in installed packages
        flow_dep = ConcreteUnmanagedVcsDependencyFlow(
            unmanaged=None, namespace_inject="test_ns"
        )
        org_config = mock.Mock()
        org_config.installed_packages = {}
        assert flow_dep._get_unmanaged(org_config) is True

        # Test with unmanaged = None and namespace_inject present and in installed packages
        flow_dep = ConcreteUnmanagedVcsDependencyFlow(
            unmanaged=None, namespace_inject="test_ns"
        )
        org_config = mock.Mock()
        org_config.installed_packages = {"test_ns": mock.Mock()}
        assert flow_dep._get_unmanaged(org_config) is False

        # Test with explicit unmanaged = True
        flow_dep = ConcreteUnmanagedVcsDependencyFlow(unmanaged=True)
        org_config = mock.Mock()
        assert flow_dep._get_unmanaged(org_config) is True

        # Test with explicit unmanaged = False
        flow_dep = ConcreteUnmanagedVcsDependencyFlow(unmanaged=False)
        org_config = mock.Mock()
        assert flow_dep._get_unmanaged(org_config) is False

    @mock.patch("cumulusci.core.dependencies.base.datetime")
    @mock.patch("cumulusci.core.dependencies.base.format_duration")
    @mock.patch("cumulusci.core.dependencies.base.FlowCoordinator")
    @mock.patch("cumulusci.vcs.vcs_source.VCSSource")
    @mock.patch("cumulusci.utils.yaml.cumulusci_yml.VCSSourceModel")
    def test_install(
        self,
        vcs_source_model_mock,
        vcs_source_mock,
        flow_coordinator_mock,
        format_duration_mock,
        datetime_mock,
    ):
        """Test the install method"""
        # Setup mocks
        mock_context = mock.Mock()
        mock_context.logger = mock.Mock()
        mock_context.allow_remote_code = False
        mock_context.keychain = mock.Mock()

        mock_org = mock.Mock()
        mock_org.installed_packages = {}

        # Setup VCS source mock
        mock_vcs_source_instance = mock.Mock()
        mock_vcs_source_instance.allow_remote_code = False
        vcs_source_mock.create.return_value = mock_vcs_source_instance

        # Setup project config mock
        mock_project_config = mock.Mock()
        mock_vcs_source_instance.fetch.return_value = mock_project_config

        # Setup flow config mock
        mock_flow_config = mock.Mock()
        mock_project_config.get_flow.return_value = mock_flow_config
        mock_project_config.list_tasks.return_value = []

        # Setup coordinator mock
        mock_coordinator_instance = mock.Mock()
        flow_coordinator_mock.return_value = mock_coordinator_instance

        # Setup datetime mock
        from datetime import datetime as real_datetime

        mock_start_time = real_datetime(2023, 1, 1, 10, 0, 0)
        mock_end_time = real_datetime(2023, 1, 1, 10, 0, 5)
        datetime_mock.now.side_effect = [mock_start_time, mock_end_time]

        # Setup format duration mock
        format_duration_mock.return_value = "5.2s"

        # Create flow dependency
        flow_dep = ConcreteUnmanagedVcsDependencyFlow(
            vcs="github",
            url="https://github.com/test/repo",
            commit="abc123",
            flow_name="install_deps",
            namespace_inject="test_ns",
            namespace_strip="old_ns",
        )

        # Call install
        flow_dep.install(mock_context, mock_org)

        # Verify VCSSourceModel was created correctly
        vcs_source_model_mock.assert_called_once_with(
            vcs="github",
            url="https://github.com/test/repo",
            commit="abc123",
            allow_remote_code=False,
        )

        # Verify VCSSource.create was called
        vcs_source_mock.create.assert_called_once_with(
            mock_context, vcs_source_model_mock.return_value
        )

        # Verify fetch was called
        mock_vcs_source_instance.fetch.assert_called_once()

        # Verify project config setup
        mock_project_config.set_keychain.assert_called_once_with(mock_context.keychain)
        assert mock_project_config.source == mock_vcs_source_instance

        # Verify flow config setup
        mock_project_config.get_flow.assert_called_once_with("install_deps")
        assert mock_flow_config.name == "install_deps"

        # Verify FlowCoordinator was created correctly
        flow_coordinator_mock.assert_called_once()
        args, kwargs = flow_coordinator_mock.call_args
        assert args[0] == mock_project_config
        assert args[1] == mock_flow_config
        assert kwargs["name"] == "install_deps"
        assert kwargs["options"] == {}
        assert kwargs["skip"] is None
        assert isinstance(kwargs["callbacks"], type(flow_dep.callback_class()))

        # Verify coordinator.run was called
        mock_coordinator_instance.run.assert_called_once_with(mock_org)

        # Verify logging
        assert (
            mock_context.logger.info.call_count == 2
        )  # Initial log, fetching log, final log
        mock_context.logger.info.assert_any_call(
            "Deploying dependency Flow from https://github.com/test/repo Flow: install_deps @abc123"
        )
        mock_context.logger.info.assert_any_call("Ran install_deps in 5.2s")

        # Verify datetime and format_duration calls
        assert datetime_mock.now.call_count == 2
        format_duration_mock.assert_called_once_with(mock_end_time - mock_start_time)


class TestParseDependency:
    def test_parse_managed_package_dep(self):
        m = parse_dependency({"version": "1.0", "namespace": "foo"})

        assert isinstance(m, PackageNamespaceVersionDependency)

        m = parse_dependency({"version_id": "04t000000000000"})

        assert isinstance(m, PackageVersionIdDependency)

    def test_parse_github_dependency(self):
        g = parse_dependency({"github": "https://github.com/Test/TestRepo"})
        assert isinstance(g, GitHubDynamicDependency)

        g = parse_dependency({"repo_owner": "Test", "repo_name": "TestRepo"})
        assert isinstance(g, GitHubDynamicDependency)

    def test_parse_unmanaged_dependency(self):
        u = parse_dependency(
            {"repo_owner": "Test", "repo_name": "TestRepo", "ref": "aaaaaaaa"}
        )
        assert isinstance(u, UnmanagedGitHubRefDependency)

        u = parse_dependency(
            {"github": "https://github.com/Test/TestRepo", "ref": "aaaaaaaa"}
        )
        assert isinstance(u, UnmanagedGitHubRefDependency)

        u = parse_dependency(
            {
                "zip_url": "https://github.com/Test/TestRepo",
                "subfolder": "unpackaged/pre",
            }
        )
        assert isinstance(u, UnmanagedZipURLDependency)

        u = parse_dependency(
            {
                "github": "https://github.com/Test/TestRepo",
                "ref": "aaaaaaaa",
                "namespace_inject": "ns",
                "collision_check": "false",
            }
        )
        assert isinstance(u, UnmanagedGitHubRefDependency)



import logging
from abc import ABC, abstractmethod
from datetime import datetime
from io import StringIO
from re import Pattern
from typing import Optional, Union


class AbstractRepo(ABC):
    """Abstract base class for repositories.
    This is the abstract base class for defining repository interfaces.

    Raises:
        NotImplementedError: Raised when a subclass does not implement the required method.
    """

    repo: object
    logger: logging.Logger
    options: dict
    repo_name: str
    repo_owner: str
    repo_url: str
    project_config: Optional[object] = None

    def __init__(self, **kwargs) -> None:
        """Initializes the AbstractRepo."""
        self.repo = kwargs.get("repo")
        self.logger = kwargs.get("logger", None)
        self.options = kwargs.get("options", {})
        self.repo_name = kwargs.get("repo_name", None)
        self.repo_owner = kwargs.get("repo_owner", None)
        self.repo_url = self.options.get("repository_url", None)

    def __eq__(self, other):
        return isinstance(other, AbstractRepo) and self.repo_url == other.repo_url

    def __hash__(self):
        return hash(self.repo_url)

    @property
    @abstractmethod
    def owner_login(self) -> str:
        """Returns the login of the repository owner."""
        raise NotImplementedError("Subclasses should provide their own implementation")

    @abstractmethod
    def get_ref(self, ref_sha: str) -> "AbstractRef":
        """Returns a Reference object for the given SHA."""
        raise NotImplementedError("Subclasses should provide their own implementation")

    @abstractmethod
    def get_ref_for_tag(self, tag_name: str) -> "AbstractRef":
        """Gets a Reference object for the tag with the given name.
        This method should be overridden by subclasses to provide
        the specific implementation for retrieving the reference for a tag.
        The method should return an instance of a class that implements
        the AbstractRef interface."""
        raise NotImplementedError("Subclasses should provide their own implementation")

    @abstractmethod
    def get_tag_by_ref(self, ref: object, tag_name: str = None) -> "AbstractGitTag":
        """Gets a Git Tag for with the given ref.
        This method should be overridden by subclasses to provide
        the specific implementation for retrieving the tag.
        The method should return an instance of a class that implements
        the AbstractGitTag interface."""
        raise NotImplementedError("Subclasses should provide their own implementation")

    @abstractmethod
    def create_tag(
        self,
        tag_name: str,
        message: str,
        sha: str,
        obj_type: str,
        tagger={},
        lightweight: Optional[bool] = False,
    ) -> "AbstractGitTag":
        """Create a Git Tag object for the tag with the given name.
        This method should be overridden by subclasses to provide
        the specific implementation for creating a Git Tag.
        The method should return an instance of a class that implements
        the AbstractGitTag interface."""
        raise NotImplementedError("Subclasses should provide their own implementation")

    @abstractmethod
    def branch(self, branch_name: str) -> "AbstractBranch":
        """Gets a Reference object for the branch with the given name.
        This method should be overridden by subclasses to provide
        the specific implementation for retrieving the reference for a branch.
        The method should return an instance of a class that implements
        the AbstractBranch interface."""
        raise NotImplementedError("Subclasses should provide their own implementation")

    @abstractmethod
    def branches(self) -> list:
        """Gets a list of all branches in the repository.
        This method should be overridden by subclasses to provide
        the specific implementation for retrieving the branches.
        The method should return a list of instances of classes that implement
        the AbstractBranch interface."""
        raise NotImplementedError("Subclasses should provide their own implementation")

    @abstractmethod
    def compare_commits(
        self, base: str, head: str, source: str
    ) -> "AbstractComparison":
        """Compares the given head with the given base.
        This method should be overridden by subclasses to provide
        the specific implementation for comparing commits.
        The method should return an instance of AbstractComparison."""
        raise NotImplementedError("Subclasses should provide their own implementation")

    @abstractmethod
    def merge(
        self, base: str, head: str, source: str, message: str = ""
    ) -> "AbstractRepoCommit":
        """Merges the given base and head with the specified commit.
        This method should be overridden by subclasses to provide
        the specific implementation for merging commits.
        The method should return an instance of AbstractRepoCommit."""
        raise NotImplementedError("Subclasses should provide their own implementation")

    @abstractmethod
    def pull_requests(self, **kwargs) -> list["AbstractPullRequest"]:
        """Fetches all pull requests from the repository.
        This method should be overridden by subclasses to provide
        the specific implementation for retrieving pull requests.
        The method should return a list of instances of classes that implement
        the AbstractPullRequest interface."""
        raise NotImplementedError("Subclasses should provide their own implementation")

    @abstractmethod
    def create_pull(
        self,
        title: str,
        base: str,
        head: str,
        body: str = None,
        maintainer_can_modify: bool = None,
        options: dict = {},
    ) -> "AbstractPullRequest":
        """Creates a pull request in the repository.
        This method should be overridden by subclasses to provide
        the specific implementation for creating pull requests.
        The method should return an instance of a class that implements
        the AbstractPullRequest interface."""
        raise NotImplementedError("Subclasses should provide their own implementation")

    @abstractmethod
    def get_commit(self, commit_sha: str) -> "AbstractRepoCommit":
        """Gets a commit object for the given commit SHA.
        This method should be overridden by subclasses to provide
        the specific implementation for retrieving the commit.
        The method should return an instance of a class that implements
        the AbstractRepoCommit interface."""
        raise NotImplementedError("Subclasses should provide their own implementation")

    @property
    @abstractmethod
    def default_branch(self) -> str:
        """Gets the default branch of the repository.
        This method should be overridden by subclasses to provide
        the specific implementation for retrieving the default branch.
        The method should return the name of the default branch."""
        raise NotImplementedError("Subclasses should provide their own implementation")

    @abstractmethod
    def archive(self, format: str, zip_content: Union[str, object], ref=None) -> bytes:
        """Archives the repository content as a zip file."""
        # This method should be overridden by subclasses to provide
        # the specific implementation for archiving the repository.
        # The method should return a bytes object representing the archived content.
        raise NotImplementedError("Subclasses should provide their own implementation")

    @abstractmethod
    def full_name(self) -> str:
        """Gets the full name of the repository.
        This method should be overridden by subclasses to provide
        the specific implementation for retrieving the full name.
        The method should return a string representing the full name of the repository."""
        raise NotImplementedError("Subclasses should provide their own implementation")

    @abstractmethod
    def release_from_tag(self, tag_name: str) -> "AbstractRelease":
        """Gets a Release object for the tag with the given name.
        This method should be overridden by subclasses to provide
        the specific implementation for retrieving the release.
        The method should return an instance of a class that implements
        the AbstractRelease interface."""
        raise NotImplementedError("Subclasses should provide their own implementation")

    @abstractmethod
    def create_release(
        self,
        tag_name: str,
        name: str,
        body: str = None,
        draft: bool = False,
        prerelease: bool = False,
        options: dict = {},
    ) -> "AbstractRelease":
        """Creates a release in the repository.
        This method should be overridden by subclasses to provide
        the specific implementation for creating releases.
        The method should return an instance of a class that implements
        the AbstractRelease interface."""
        raise NotImplementedError("Subclasses should provide their own implementation")

    @abstractmethod
    def releases(self) -> list["AbstractRelease"]:
        """Fetches all releases from the repository.
        This method should be overridden by subclasses to provide
        the specific implementation for retrieving releases.
        The method should return a list of instances of classes that implement
        the AbstractRelease interface."""
        raise NotImplementedError("Subclasses should provide their own implementation")

    @abstractmethod
    def latest_release(self) -> Optional["AbstractRelease"]:
        """Gets the latest release from the repository.
        This method should be overridden by subclasses to provide
        the specific implementation for retrieving the latest release.
        The method should return an instance of a class that implements
        the AbstractRelease interface."""
        raise NotImplementedError("Subclasses should provide their own implementation")

    @abstractmethod
    def has_issues(self) -> bool:
        """Checks if the repository supports issues.
        This method should be overridden by subclasses to provide
        the specific implementation for checking if the repository supports issues.
        The method should return True if the repository supports issues, otherwise False."""
        raise NotImplementedError(
            "Subclasses should implement the has_issues method to check if the repository supports issues."
        )

    @abstractmethod
    def get_pr_issue_labels(self, pull_request: "AbstractPullRequest") -> list[str]:
        """Gets the issue labels for the given pull request.
        This method should be overridden by subclasses to provide
        the specific implementation for retrieving the labels.
        The method should return a list of labels associated with the pull request."""
        raise NotImplementedError("Subclasses should implement this method.")

    @abstractmethod
    def get_latest_prerelease(self) -> Optional["AbstractRelease"]:
        """Gets the latest pre-release from the repository.
        This method should be overridden by subclasses to provide
        the specific implementation for retrieving the latest pre-release.
        The method should return an instance of a class that implements
        the AbstractRelease interface."""
        raise NotImplementedError("Subclasses should implement this method.")

    @abstractmethod
    def directory_contents(self, subfolder: str, return_as, ref: str) -> dict:
        """Gets the contents of a directory in the repository.
        This method should be overridden by subclasses to provide
        the specific implementation for retrieving the directory contents.
        The method should return the contents of the directory as a dictionary or list,
        depending on the value of return_as."""
        raise NotImplementedError("Subclasses should implement this method.")

    @abstractmethod
    def file_contents(self, file: str, ref: str) -> StringIO:
        """Gets the contents of a file in the repository.
        This method should be overridden by subclasses to provide
        the specific implementation for retrieving the file contents.
        The method should return the contents of the file as a StringIO object."""
        raise NotImplementedError("Subclasses should implement this method.")

    @property
    @abstractmethod
    def clone_url(self) -> str:
        """Gets the https clone URL for the repository.
        This method should be overridden by subclasses to provide
        the specific implementation for retrieving the clone URL.
        The method should return a string representing the clone URL of the repository."""
        raise NotImplementedError("Subclasses should implement this method.")

    @abstractmethod
    def create_commit_status(
        self,
        commit_id: str,
        context: str,
        state: str,
        description: str,
        target_url: str,
    ) -> "AbstractRepoCommit":
        """Creates a commit status in the repository.
        This method should be overridden by subclasses to provide
        the specific implementation for creating commit statuses.
        The method should not return anything, but it should raise an exception
        if the commit status creation fails."""
        raise NotImplementedError("Subclasses should implement this method.")


class AbstractRelease(ABC):
    """Abstract base class for releases.
    This is the abstract base class for defining git releases.

    Raises:
        NotImplementedError: Raised when a subclass does not implement the required method.
    """

    release: object

    def __init__(self, **kwargs) -> None:
        """Initializes the AbstractRelease."""
        self.release = kwargs.get("release", None)

    @property
    @abstractmethod
    def tag_name(self) -> str:
        """Gets the tag name of the release."""
        raise NotImplementedError("Subclasses should implement this method.")

    @property
    @abstractmethod
    def body(self) -> str:
        """Gets the body of the release."""
        raise NotImplementedError("Subclasses should implement this method.")

    @property
    @abstractmethod
    def prerelease(self) -> bool:
        """Checks if the release is a pre-release."""
        raise NotImplementedError("Subclasses should implement this method.")

    @property
    @abstractmethod
    def name(self) -> str:
        """Gets the name of the release."""
        raise NotImplementedError("Subclasses should implement this method.")

    @property
    @abstractmethod
    def html_url(self) -> str:
        """Gets the HTML URL of the release."""
        raise NotImplementedError("Subclasses should implement this method.")

    @property
    @abstractmethod
    def created_at(self) -> datetime:
        """Gets the creation date of the release."""
        raise NotImplementedError("Subclasses should implement this method.")

    @property
    @abstractmethod
    def draft(self) -> bool:
        """Checks if the release is a draft."""
        raise NotImplementedError("Subclasses should implement this method.")

    @property
    @abstractmethod
    def tag_ref_name(self) -> str:
        """Gets the tag reference name of the release."""
        raise NotImplementedError("Subclasses should implement this method.")

    @property
    def updateable(self) -> bool:
        """Checks if the release is updateable."""
        return False


class AbstractRef(ABC):
    """Abstract base class for Refs.
    This is the abstract base class for defining git references.

    Raises:
        NotImplementedError: Raised when a subclass does not implement the required method.
    """

    ref: object

    def __init__(self, ref, **kwargs) -> None:
        """Initializes the AbstractRef."""
        self.ref = ref


class AbstractGitTag(ABC):

    tag: object
    _sha: str = ""

    def __init__(self, **kwargs) -> None:
        """Initializes the AbstractGitTag."""
        self.tag = kwargs.get("tag", None)

    @property
    def sha(self) -> str:
        """Gets the SHA for the git tag."""
        return self._sha

    @sha.setter
    def sha(self, value: str):
        """Sets the SHA for the git tag.

        Args:
            value (str): The SHA for the git tag.
        """
        self._sha = value

    @property
    @abstractmethod
    def message(self) -> str:
        """Gets the message of the tag."""
        raise NotImplementedError("Subclasses should implement this method.")


class AbstractBranch(ABC):
    """Abstract base class for Branch.
    This is the abstract base class for defining git branches.

    Raises:
        NotImplementedError: Raised when a subclass does not implement the required method.
    """

    repo: AbstractRepo
    branch: object
    name: str

    def __init__(self, repo: AbstractRepo, branch_name: str, **kwargs) -> None:
        """Initializes the AbstractBranch."""
        self.repo = repo
        self.name = branch_name

        self.branch = kwargs.get("branch", None)
        if not self.branch:
            self.get_branch()

    @abstractmethod
    def get_branch(self) -> None:
        """Gets the branch object for the current branch name."""
        raise NotImplementedError("Subclasses should implement this method.")

    @classmethod
    @abstractmethod
    def branches(cls, repo: AbstractRepo) -> list["AbstractBranch"]:
        """Gets a list of all branches in the repository.
        This method should be overridden by subclasses to provide
        the specific implementation for retrieving the branches.
        The method should return a list of instances of classes that implement
        the AbstractBranch interface."""
        raise NotImplementedError("Subclasses should provide their own implementation")

    @property
    @abstractmethod
    def commit(self) -> "AbstractRepoCommit":
        """Gets the branch commit for the current branch."""
        raise NotImplementedError("Subclasses should provide their own implementation")


class AbstractComparison(ABC):
    """Abstract base class for comparisons.
    This is the abstract base class for defining git comparisons.

    Raises:
        NotImplementedError: Raised when a subclass does not implement the required method.
    """

    comparison: object

    def __init__(self, repo: AbstractRepo, base: str, head: str, **kwargs) -> None:
        """Initializes the AbstractComparison."""
        self.repo = repo
        self.base = base
        self.head = head
        self.comparison = kwargs.get("comparison", None)

    @classmethod
    @abstractmethod
    def compare(cls, repo: AbstractRepo, base: str, head: str) -> "AbstractComparison":
        """Compares the given base and head with the specified commit.
        This method should be overridden by subclasses to provide
        the specific implementation for comparing commits.
        The method should return an instance of AbstractComparison."""
        raise NotImplementedError("Subclasses should provide their own implementation")

    @abstractmethod
    def get_comparison(self) -> None:
        """Gets the comparison object for the current base and head."""
        raise NotImplementedError("Subclasses should implement this method.")

    @property
    def files(self) -> list:
        """Gets the files changed in the comparison."""
        raise NotImplementedError("Subclasses should implement this method.")

    @property
    def behind_by(self) -> int:
        """Gets the number of commits the head is behind the base."""
        raise NotImplementedError("Subclasses should implement this method.")


class AbstractRepoCommit(ABC):
    """Abstract base class for commits.
    This is the abstract base class for defining git commits.

    Raises:
        NotImplementedError: Raised when a subclass does not implement the required method.
    """

    commit: object

    def __init__(self, **kwargs) -> None:
        """Initializes the AbstractRepoCommit."""
        self.commit = kwargs.get("commit", None)

    @abstractmethod
    def get_statuses(self, context: str, regex_match: Pattern[str]) -> Optional[str]:
        """Gets the statuses for the commit.
        This method should be overridden by subclasses to provide
        the specific implementation for retrieving the statuses.
        The method should return a string if a match is found, otherwise None."""
        raise NotImplementedError("Subclasses should implement this method.")

    @property
    @abstractmethod
    def parents(self) -> list["AbstractRepoCommit"]:
        raise NotImplementedError("Subclasses should implement this method.")

    @property
    @abstractmethod
    def sha(self) -> str:
        """Gets the SHA of the commit."""
        raise NotImplementedError("Subclasses should implement this method.")


class AbstractPullRequest(ABC):
    """Abstract base class for pull requests.
    This is the abstract base class for defining git pull requests.

    Raises:
        NotImplementedError: Raised when a subclass does not implement the required method.
    """

    repo: Union[AbstractRepo, None]
    pull_request: Union[object, None]

    def __init__(self, **kwargs) -> None:
        """Initializes the AbstractPullRequest."""
        self.repo = kwargs.get("repo", None)
        self.pull_request = kwargs.get("pull_request", None)

    @classmethod
    @abstractmethod
    def pull_requests(cls, *args, **kwargs) -> list["AbstractPullRequest"]:
        """Fetches all pull requests from the repository."""
        raise NotImplementedError("Subclasses should implement this method.")

    @classmethod
    @abstractmethod
    def create_pull(cls, *args, **kwargs) -> "AbstractPullRequest":
        """Creates a pull request on the repository."""
        raise NotImplementedError("Subclasses should implement this method.")

    @property
    @abstractmethod
    def number(self) -> int:
        """Gets the pull request number."""
        raise NotImplementedError("Subclasses should implement this method.")

    @property
    @abstractmethod
    def title(self) -> str:
        """Gets the pull title."""
        raise NotImplementedError("Subclasses should implement this method.")

    @property
    @abstractmethod
    def base_ref(self) -> str:
        """Gets the base reference of the pull request."""
        raise NotImplementedError("Subclasses should implement this method.")

    @property
    @abstractmethod
    def head_ref(self) -> str:
        """Gets the head reference of the pull request."""
        raise NotImplementedError("Subclasses should implement this method.")

    @property
    @abstractmethod
    def merged_at(self) -> datetime:
        """Gets the date when the pull request was merged."""
        raise NotImplementedError("Subclasses should implement this method.")



from unittest.mock import ANY, MagicMock, Mock, mock_open, patch

import pytest

from cumulusci.core.config import TaskConfig
from cumulusci.core.exceptions import TaskOptionsError
from cumulusci.tasks.salesforce.check_components import CheckComponents
from cumulusci.tests.util import create_project_config

from .util import create_task


class TestCheckComponents:
    @patch("os.path.exists", return_value=True)
    @patch("os.remove")
    @patch("os.path.isdir", return_value=True)
    @patch("os.listdir", return_value=["package.xml"])
    @patch("os.path.join", side_effect=lambda *args: "/".join(args))
    @patch("cumulusci.core.sfdx.convert_sfdx_source")
    @patch(
        "cumulusci.tasks.salesforce.check_components.CheckComponents._is_plan",
        return_value=False,
    )
    @patch(
        "cumulusci.tasks.salesforce.check_components.CheckComponents._freeze_steps",
        return_value=[],
    )
    @patch(
        "cumulusci.tasks.salesforce.check_components.CheckComponents._collect_components_from_paths",
        return_value=[{"Type1": ["Comp1"]}, []],
    )
    @patch(
        "cumulusci.tasks.salesforce.check_components.CheckComponents._get_api_object_responce",
        return_value=[],
    )
    @patch(
        "builtins.open",
        new_callable=mock_open,
        read_data="""
        <Package xmlns="http://soap.sforce.com/2006/04/metadata">
            <types>
                <members>Delivery</members>
                <name>ApexClass</name>
            </types>
            <types>
                <members>Delivery__c</members>
                <name>CustomObject</name>
            </types>
            <version>58.0</version>
        </Package>
    """,
    )
    @patch("cumulusci.utils.xml.metadata_tree.parse")
    @patch(
        "cumulusci.utils.xml.metadata_tree.parse_package_xml_types",
        return_value={"Type2": ["Comp2"]},
    )
    def test_get_repo_existing_components(
        self,
        mock_metadata_parse,
        mock_open_file,
        mock_convert_sfdx_source,
        mock_path_join,
        mock_listdir,
        mock_isdir,
        mock_remove,
        mock_path_exists,
        mock_is_plan,
        mock_freeze_steps,
        mock_get_api_object,
        mock_parse,
        mock_collect_components,
    ):
        org_config = Mock(scratch=True, config={})
        org_config.username = "test_user"
        org_config.org_id = "test_org_id"
        self.org_config = Mock(return_value=("test", org_config))
        project_config = create_project_config()
        flow_config = {
            "test": {
                "steps": {
                    1: {
                        "flow": "test2",
                    }
                }
            },
            "test2": {
                "steps": {
                    1: {
                        "task": "deploy",
                        "options": {"path": "force-app/main/default"},
                    }
                }
            },
        }
        plan_config = {
            "title": "Test Install",
            "slug": "install",
            "tier": "primary",
            "steps": {1: {"flow": "test"}},
        }
        project_config.config["plans"] = {
            "Test Install": plan_config,
        }
        project_config.config["flows"] = flow_config

        task = create_task(CheckComponents, {"name": "test2"})
        task.deploy_paths = ["test"]

        (components, response_messages) = task.get_repo_existing_components("test2")
        assert "Type1" in components
        assert "Type2" in components
        assert "Comp1" in components["Type1"]
        assert "Comp2" in components["Type2"]

    @patch("os.path.exists", return_value=True)
    @patch("os.remove")
    @patch("os.path.isdir", return_value=True)
    @patch("os.listdir", return_value=["package.xml"])
    @patch("os.path.join", side_effect=lambda *args: "/".join(args))
    @patch("cumulusci.core.sfdx.convert_sfdx_source")
    @patch(
        "cumulusci.tasks.salesforce.check_components.CheckComponents._is_plan",
        return_value=False,
    )
    @patch(
        "cumulusci.tasks.salesforce.check_components.CheckComponents._freeze_steps",
        return_value=[],
    )
    @patch(
        "cumulusci.tasks.salesforce.check_components.CheckComponents._collect_components_from_paths",
        return_value=[{"Type1": ["Comp1"]}, []],
    )
    @patch(
        "cumulusci.tasks.salesforce.check_components.CheckComponents._get_api_object_responce",
        return_value=[],
    )
    @patch(
        "builtins.open",
        new_callable=mock_open,
        read_data="""
        <Package xmlns="http://soap.sforce.com/2006/04/metadata">
            <types>
                <members>Delivery</members>
                <name>ApexClass</name>
            </types>
            <types>
                <members>Delivery__c</members>
                <name>CustomObject</name>
            </types>
            <version>58.0</version>
        </Package>
    """,
    )
    @patch("cumulusci.utils.xml.metadata_tree.parse")
    @patch(
        "cumulusci.utils.xml.metadata_tree.parse_package_xml_types",
        return_value={"Type2": ["Comp2"]},
    )
    def test_get_repo_existing_components_paths_paramter(
        self,
        mock_metadata_parse,
        mock_open_file,
        mock_convert_sfdx_source,
        mock_path_join,
        mock_listdir,
        mock_isdir,
        mock_remove,
        mock_path_exists,
        mock_is_plan,
        mock_freeze_steps,
        mock_get_api_object,
        mock_parse,
        mock_collect_components,
    ):
        org_config = Mock(scratch=True, config={})
        org_config.username = "test_user"
        org_config.org_id = "test_org_id"
        self.org_config = Mock(return_value=("test", org_config))
        project_config = create_project_config()
        flow_config = {
            "test": {
                "steps": {
                    1: {
                        "flow": "test2",
                    }
                }
            },
            "test2": {
                "steps": {
                    1: {
                        "task": "deploy",
                        "options": {"path": "force-app/main/default"},
                    }
                }
            },
        }
        plan_config = {
            "title": "Test Install",
            "slug": "install",
            "tier": "primary",
            "steps": {1: {"flow": "test"}},
        }
        project_config.config["plans"] = {
            "Test Install": plan_config,
        }
        project_config.config["flows"] = flow_config

        task = create_task(CheckComponents, {"name": "test2"})
        task.deploy_paths = ["test"]

        (components, response_messages) = task.get_repo_existing_components("", "src")
        assert "Type1" in components
        assert "Type2" in components
        assert "Comp1" in components["Type1"]
        assert "Comp2" in components["Type2"]

    @patch("os.path.exists", return_value=True)
    @patch("os.remove")
    @patch("os.path.isdir", return_value=True)
    @patch("os.listdir", return_value=["some_file_or_directory"])
    @patch("os.path.join", side_effect=lambda *args: "/".join(args))
    @patch("cumulusci.core.sfdx.convert_sfdx_source")
    @patch(
        "builtins.open",
        new_callable=mock_open,
        read_data="""
        <Package xmlns="http://soap.sforce.com/2006/04/metadata">
            <types>
                <members>Delivery</members>
                <name>ApexClass</name>
            </types>
            <types>
                <members>Delivery__c</members>
                <name>CustomObject</name>
            </types>
            <version>58.0</version>
        </Package>
    """,
    )
    @patch("cumulusci.utils.xml.metadata_tree.parse")
    def test_collect_components_from_paths(
        self,
        mock_metadata_parse,
        mock_open_file,
        mock_convert_sfdx_source,
        mock_path_join,
        mock_listdir,
        mock_isdir,
        mock_remove,
        mock_path_exists,
    ):
        mock_path_exists.return_value = True
        mock_isdir.return_value = True
        mock_listdir.return_value = ["some_file_or_directory"]
        mock_path_join.side_effect = lambda *args: "/".join(args)
        mock_convert_sfdx_source.return_value.__enter__.return_value = "/converted/path"

        mock_tree = MagicMock()
        mock_tree.findall.return_value = [
            MagicMock(
                findall=lambda tag: [MagicMock(text="Delivery")]
                if tag == "members"
                else [],
                find=lambda tag: MagicMock(text="ApexClass") if tag == "name" else None,
            ),
            MagicMock(
                findall=lambda tag: [MagicMock(text="Delivery__c")]
                if tag == "members"
                else [],
                find=lambda tag: MagicMock(text="CustomObject")
                if tag == "name"
                else None,
            ),
        ]
        mock_tree.find.return_value = MagicMock(text="58.0")
        mock_metadata_parse.return_value = mock_tree

        response_messages = [
            MagicMock(
                getElementsByTagName=MagicMock(
                    return_value=[
                        MagicMock(
                            firstChild=MagicMock(
                                nodeValue="Entity of type 'ApexClass' named 'CustomHealth' is cannot be found"
                            )
                        )
                    ]
                )
            )
        ]

        with patch("cumulusci.core.sfdx.sfdx") as sfdx:
            with patch.object(
                CheckComponents, "_get_api_object_responce"
            ) as mock_get_api_response:
                mock_get_api_response.return_value = response_messages
                task = create_task(CheckComponents, {"paths": "force-app/main/default"})
                components, api_response = task._collect_components_from_paths(
                    "force-app/main/default"
                )

                assert components is not None
                assert "ApexClass" not in components
        sfdx.assert_called_once_with(
            "project convert source",
            args=["-d", ANY, "-r", "force-app/main/default"],
            capture_output=True,
            check_return=True,
        )

    @patch("os.path.exists", return_value=False)
    def test_collect_components_from_nonexistent_paths(self, mock_path_exists):
        task = create_task(CheckComponents, {"paths": "invalid/path"})
        components, api_response = task._collect_components_from_paths("invalid/path")
        assert components is None
        assert api_response is None

    @patch("os.path.exists", return_value=False)
    def test_copy_to_tempdir_nonexistent_src(self, mock_exists):
        task = create_task(CheckComponents, {"paths": "force-app/main/default"})
        with pytest.raises(FileNotFoundError):
            task._copy_to_tempdir("nonexistent_src", "temp_dir")

    @patch("shutil.copy2")
    @patch("os.listdir", return_value=["file1", "file2"])
    @patch("os.path.isdir", return_value=False)
    @patch("os.path.exists", return_value=False)
    def test_copy_to_tempdir(self, mock_exists, mock_isdir, mock_listdir, mock_copy2):
        task = create_task(CheckComponents, {"paths": "force-app/main/default"})
        task._copy_to_tempdir("force-app/main/default", "temp_dir")
        mock_copy2.assert_called()

    def test_get_deployable_paths(self):
        task = create_task(CheckComponents, {"paths": "force-app/main/default"})
        tasks = [
            {
                "task_config": {
                    "options": {
                        "path": "unpackaged/pre",
                        "subfolder": "/unpackaged/post/test",
                    }
                }
            },
            {"task_config": {"options": {"path": "force-app"}}},
        ]
        paths = task._get_deployable_paths(tasks)
        assert paths == ["unpackaged/pre", "/unpackaged/post/test", "force-app"]

    def test_is_plan_valid(self):
        task = create_task(CheckComponents, {"paths": "force-app/main/default"})
        with patch.object(task.project_config, "lookup") as mock_lookup:
            mock_lookup.side_effect = lambda name: {
                "plans__test_plan": True,
                "flows__test_flow": False,
            }.get(name, None)
            assert task._is_plan("test_plan") is True
            assert task._is_plan("test_flow") is False

    def test_is_plan_invalid(self):
        task = create_task(CheckComponents, {"name": "invalid_name"})
        with patch.object(task.project_config, "lookup", return_value=None):
            with pytest.raises(
                TaskOptionsError,
                match="No paths provided and unable to determine the current plan or flow name.",
            ):
                task._is_plan("invalid_name")

    def test_init_options_with_both_paths_and_name(self):
        with pytest.raises(
            TaskOptionsError, match="Please provide either --paths or --name"
        ):
            create_task(CheckComponents, {"paths": "some/path", "name": "some_plan"})

    def test_init_options_with_neither_paths_nor_name(self):
        with pytest.raises(
            TaskOptionsError,
            match="This task requires a plan/flow name or paths options. Pass --paths or --name options",
        ):
            create_task(CheckComponents, {})

    def test_load_deploy_paths(self):
        task = create_task(CheckComponents, {"name": "some_name"})
        with patch.object(
            task,
            "_get_plan_tasks",
            return_value=[{"task_config": {"options": {"path": "unpackaged/pre"}}}],
        ):
            with patch.object(
                task, "_get_deployable_paths", return_value=["unpackaged/pre"]
            ):
                task._load_deploy_paths("some_name", is_plan=True)
                assert task.deploy_paths == ["unpackaged/pre"]

    def test_freeze_steps__skip(self):
        project_config = create_project_config()
        plan_config = {
            "title": "Test Install",
            "slug": "install",
            "tier": "primary",
            "steps": {1: {"task": "None"}},
        }
        project_config.config["plans"] = {
            "Test Install": plan_config,
        }
        task_config = TaskConfig({"options": {"name": "Test Install"}})
        task = CheckComponents(project_config, task_config)
        steps = task._freeze_steps(project_config, plan_config)
        assert steps == []

    def test_freeze_steps_nested(self):
        project_config = create_project_config()
        flow_config = {
            "test": {
                "steps": {
                    1: {
                        "flow": "test2",
                    }
                }
            },
            "test2": {
                "steps": {
                    1: {
                        "task": "deploy",
                        "options": {"path": "force-app/main/default"},
                    }
                }
            },
        }
        plan_config = {
            "title": "Test Install",
            "slug": "install",
            "tier": "primary",
            "steps": {1: {"flow": "test"}},
        }
        project_config.config["plans"] = {
            "Test Install": plan_config,
        }
        project_config.config["flows"] = flow_config

        task_config = TaskConfig({"options": {"name": "Test Install"}})
        task = CheckComponents(project_config, task_config)
        steps = task._freeze_steps(project_config, plan_config)
        assert steps[0]["name"] == "deploy"
        assert len(steps) == 1



from io import StringIO
from unittest.mock import MagicMock, Mock, patch

import pytest

from cumulusci.core.config import BaseProjectConfig
from cumulusci.core.exceptions import VcsException, VcsNotFoundError
from cumulusci.core.keychain import EncryptedFileProjectKeychain
from cumulusci.vcs import bootstrap
from cumulusci.vcs.base import VCSService
from cumulusci.vcs.models import (
    AbstractBranch,
    AbstractGitTag,
    AbstractPullRequest,
    AbstractRelease,
    AbstractRepo,
    AbstractRepoCommit,
)
from cumulusci.vcs.tests.dummy_service import DummyRef, DummyRepo, DummyTag


class TestBootstrapFunctions:
    """Test class for VCS bootstrap functions"""

    @pytest.fixture()
    def keychain(self, project_config, key) -> EncryptedFileProjectKeychain:
        keychain = EncryptedFileProjectKeychain(project_config, key)
        assert keychain.project_config == project_config
        assert keychain.key == key
        return keychain

    @pytest.fixture()
    def mock_repo(self):
        """Create a mock repository for testing"""
        repo = Mock(spec=AbstractRepo)
        repo.default_branch = "main"
        repo.repo_url = "https://github.com/test/repo"
        repo.owner_login = "test"
        return repo

    @pytest.fixture()
    def mock_commit(self):
        """Create a mock commit for testing"""
        commit = Mock(spec=AbstractRepoCommit)
        commit.sha = "1234567890abcdef"
        commit.parents = []
        return commit

    @pytest.fixture()
    def mock_release(self):
        """Create a mock release for testing"""
        release = Mock(spec=AbstractRelease)
        release.tag_name = "release/1.0.0"
        release.prerelease = False
        return release

    @pytest.fixture()
    def mock_pull_request(self):
        """Create a mock pull request for testing"""
        pr = Mock(spec=AbstractPullRequest)
        pr.merged_at = None
        pr.number = 123
        return pr

    @pytest.fixture()
    def mock_project_config(self):
        """Create a mock project config for testing"""
        config = Mock(spec=BaseProjectConfig)
        config.repo_url = "https://github.com/test/repo"
        config.config = {}
        config.config["services"] = {
            "test_service": {
                "attributes": {"name": {"required": True}, "password": {}},
            },
        }
        config.logger = Mock()
        config.project__git__prefix_feature = "feature/"
        config.project__git__prefix_release = "release/"
        config.project__git__prefix_beta = "beta/"
        return config

    def test_get_service_with_class_path(self, keychain, service_config):
        """Test getting service with valid class path"""
        project_config = keychain.project_config
        project_config.keychain = keychain
        keychain.set_service("github", "alias", service_config)

        encrypted = keychain._get_config_bytes(service_config)
        keychain.config["services"]["github"] = {"alias": encrypted}

        keychain._save_default_service("github", "alias", project=False)
        keychain._load_default_services()

        result = bootstrap.get_service(project_config)
        assert isinstance(result, VCSService)

    @patch("cumulusci.vcs.base.VCSService.registered_services")
    def test_get_service_missing_class_path(self, mock_registered_services):
        """Test error handling when class path is missing"""
        config = MagicMock()
        config.get_project_service.return_value = ("github", None)
        config.lookup.return_value = "github"
        config.services = {"github": {}}
        config.repo_url = "https://github.com/invalid/repo"

        # Mock no registered services to force the exception
        mock_registered_services.return_value = []

        with pytest.raises(VcsException) as e:
            bootstrap.get_service(config)
        assert "Could not find a VCS service for URL" in str(e.value)

    def test_get_ref_for_tag(self):
        """Test getting reference for a tag"""
        repo = DummyRepo()
        result = bootstrap.get_ref_for_tag(repo, "v1.0")
        assert isinstance(result, DummyRef)

    def test_get_tag_by_name(self):
        """Test getting tag by name"""
        repo = DummyRepo()
        result = bootstrap.get_tag_by_name(repo, "v1.0")
        assert isinstance(result, DummyTag)

    def test_get_version_id_from_commit_success(self, mock_repo):
        """Test successfully getting version ID from commit"""
        mock_commit = Mock(spec=AbstractRepoCommit)
        mock_commit.get_statuses.return_value = "04t123456789"
        mock_repo.get_commit.return_value = mock_commit

        result = bootstrap.get_version_id_from_commit(mock_repo, "abc123", "context")
        assert result == "04t123456789"

    def test_get_version_id_from_commit_no_commit(self, mock_repo):
        """Test getting version ID when commit doesn't exist"""
        mock_repo.get_commit.return_value = None

        result = bootstrap.get_version_id_from_commit(mock_repo, "abc123", "context")
        assert result is None

    def test_get_commit(self, mock_repo, mock_commit):
        """Test getting commit by SHA"""
        mock_repo.get_commit.return_value = mock_commit

        result = bootstrap.get_commit(mock_repo, "abc123")
        assert result == mock_commit
        mock_repo.get_commit.assert_called_once_with("abc123")

    def test_get_pull_requests_with_base_branch_simple(self, mock_repo):
        """Test getting pull requests with base branch"""
        mock_prs = [Mock(), Mock()]
        mock_repo.pull_requests.return_value = mock_prs

        result = bootstrap.get_pull_requests_with_base_branch(mock_repo, "main")
        assert result == mock_prs
        mock_repo.pull_requests.assert_called_once_with(
            base="main", head=None, state=None
        )

    def test_get_pull_requests_with_base_branch_with_head(self, mock_repo):
        """Test getting pull requests with base branch and head"""
        mock_prs = [Mock(), Mock()]
        mock_repo.pull_requests.return_value = mock_prs
        mock_repo.owner_login = "test_owner"

        result = bootstrap.get_pull_requests_with_base_branch(
            mock_repo, "main", head="feature-branch", state="open"
        )
        assert result == mock_prs
        mock_repo.pull_requests.assert_called_once_with(
            base="main", head="test_owner:feature-branch", state="open"
        )

    def test_is_pull_request_merged_true(self):
        """Test pull request is merged"""
        pr = Mock(spec=AbstractPullRequest)
        pr.merged_at = "2023-01-01T00:00:00Z"

        result = bootstrap.is_pull_request_merged(pr)
        assert result is True

    def test_is_pull_request_merged_false(self):
        """Test pull request is not merged"""
        pr = Mock(spec=AbstractPullRequest)
        pr.merged_at = None

        result = bootstrap.is_pull_request_merged(pr)
        assert result is False

    def test_is_label_on_pull_request_true(self, mock_repo, mock_pull_request):
        """Test label exists on pull request"""
        mock_repo.get_pr_issue_labels.return_value = [
            "bug",
            "enhancement",
            "documentation",
        ]

        result = bootstrap.is_label_on_pull_request(mock_repo, mock_pull_request, "bug")
        assert result is True

    def test_is_label_on_pull_request_false(self, mock_repo, mock_pull_request):
        """Test label does not exist on pull request"""
        mock_repo.get_pr_issue_labels.return_value = ["bug", "enhancement"]

        result = bootstrap.is_label_on_pull_request(
            mock_repo, mock_pull_request, "documentation"
        )
        assert result is False

    @patch("cumulusci.vcs.bootstrap.get_service_for_repo_url")
    def test_get_repo_from_url_success(self, mock_get_service, mock_project_config):
        """Test successfully getting repository from URL"""
        mock_service = Mock(spec=VCSService)
        mock_repo = Mock(spec=AbstractRepo)
        mock_service.get_repository.return_value = mock_repo
        mock_get_service.return_value = mock_service

        result = bootstrap.get_repo_from_url(
            mock_project_config, "https://github.com/test/repo"
        )

        assert result == mock_repo
        mock_get_service.assert_called_once_with(
            mock_project_config, "https://github.com/test/repo", service_alias=None
        )

    @patch("cumulusci.vcs.bootstrap.get_service_for_repo_url")
    def test_get_repo_from_url_not_found(self, mock_get_service, mock_project_config):
        """Test error when repository is not found"""
        mock_service = Mock(spec=VCSService)
        mock_service.get_repository.return_value = None
        mock_get_service.return_value = mock_service

        with pytest.raises(VcsNotFoundError) as e:
            bootstrap.get_repo_from_url(
                mock_project_config, "https://github.com/test/repo"
            )

        assert "Could not find a repository at https://github.com/test/repo" in str(
            e.value
        )

    @patch("cumulusci.vcs.base.VCSService.registered_services")
    def test_get_service_for_repo_url_success(
        self, mock_registered_services, mock_project_config
    ):
        """Test successfully getting service for repo URL"""
        mock_service_class = Mock()
        mock_service_instance = Mock(spec=VCSService)
        mock_service_class.get_service_for_url.return_value = mock_service_instance
        mock_registered_services.return_value = [mock_service_class]

        result = bootstrap.get_service_for_repo_url(
            mock_project_config, "https://github.com/test/repo"
        )

        assert result == mock_service_instance
        mock_service_class.get_service_for_url.assert_called_once_with(
            mock_project_config, "https://github.com/test/repo", service_alias=None
        )

    @patch("cumulusci.vcs.base.VCSService.registered_services")
    def test_get_service_for_repo_url_not_found(
        self, mock_registered_services, mock_project_config
    ):
        """Test error when no service found for URL"""
        mock_service_class = Mock()
        mock_service_class.get_service_for_url.return_value = None
        mock_registered_services.return_value = [mock_service_class]

        with pytest.raises(VcsException) as e:
            bootstrap.get_service_for_repo_url(
                mock_project_config, "https://github.com/test/repo"
            )

        assert (
            "Could not find a VCS service for URL: https://github.com/test/repo"
            in str(e.value)
        )

    @patch("cumulusci.vcs.bootstrap.cci_safe_load")
    def test_get_remote_project_config(self, mock_cci_load, mock_repo):
        """Test getting remote project configuration"""
        mock_file_contents = StringIO("project:\n  name: test")
        mock_repo.file_contents.return_value = mock_file_contents
        mock_cci_load.return_value = {"project": {"name": "test"}}

        result = bootstrap.get_remote_project_config(mock_repo, "main")

        assert isinstance(result, BaseProjectConfig)
        mock_repo.file_contents.assert_called_once_with("cumulusci.yml", ref="main")
        mock_cci_load.assert_called_once_with(mock_file_contents)

    @patch("cumulusci.vcs.bootstrap.get_remote_project_config")
    def test_find_repo_feature_prefix_with_custom_prefix(
        self, mock_get_config, mock_repo
    ):
        """Test finding repository feature prefix with custom prefix"""
        mock_branch = Mock()
        mock_commit = Mock()
        mock_commit.sha = "abc123"
        mock_branch.commit = mock_commit
        mock_repo.branch.return_value = mock_branch

        mock_config = Mock()
        mock_config.project__git__prefix_feature = "custom-feature/"
        mock_get_config.return_value = mock_config

        result = bootstrap.find_repo_feature_prefix(mock_repo)

        assert result == "custom-feature/"
        mock_repo.branch.assert_called_once_with(mock_repo.default_branch)

    @patch("cumulusci.vcs.bootstrap.get_remote_project_config")
    def test_find_repo_feature_prefix_default(self, mock_get_config, mock_repo):
        """Test finding repository feature prefix with default"""
        mock_branch = Mock()
        mock_commit = Mock()
        mock_commit.sha = "abc123"
        mock_branch.commit = mock_commit
        mock_repo.branch.return_value = mock_branch

        mock_config = Mock()
        mock_config.project__git__prefix_feature = None
        mock_get_config.return_value = mock_config

        result = bootstrap.find_repo_feature_prefix(mock_repo)

        assert result == "feature/"

    @patch("cumulusci.vcs.bootstrap.get_remote_project_config")
    def test_get_remote_context(self, mock_get_config, mock_repo):
        """Test getting remote context"""
        mock_config = Mock()
        mock_config.lookup.return_value = "custom_context"
        mock_get_config.return_value = mock_config

        result = bootstrap.get_remote_context(
            mock_repo, "test_context", "default_context"
        )

        assert result == "custom_context"
        mock_config.lookup.assert_called_once_with("project__git__test_context")

    @patch("cumulusci.vcs.bootstrap.get_remote_project_config")
    def test_get_remote_context_default(self, mock_get_config, mock_repo):
        """Test getting remote context with default"""
        mock_config = Mock()
        mock_config.lookup.return_value = None
        mock_get_config.return_value = mock_config

        result = bootstrap.get_remote_context(
            mock_repo, "test_context", "default_context"
        )

        assert result == "default_context"

    @patch("cumulusci.vcs.bootstrap.get_latest_prerelease")
    def test_find_latest_release_beta(self, mock_get_prerelease, mock_repo):
        """Test finding latest release including beta"""
        mock_release = Mock(spec=AbstractRelease)
        mock_get_prerelease.return_value = mock_release

        result = bootstrap.find_latest_release(mock_repo, include_beta=True)

        assert result == mock_release
        mock_get_prerelease.assert_called_once_with(mock_repo)

    def test_find_latest_release_production(self, mock_repo):
        """Test finding latest production release"""
        mock_release = Mock(spec=AbstractRelease)
        mock_repo.latest_release.return_value = mock_release

        result = bootstrap.find_latest_release(mock_repo, include_beta=False)

        assert result == mock_release
        mock_repo.latest_release.assert_called_once()

    def test_find_latest_release_not_found(self, mock_repo):
        """Test finding latest release when none exists"""
        mock_repo.latest_release.side_effect = VcsNotFoundError("No releases found")

        result = bootstrap.find_latest_release(mock_repo, include_beta=False)

        assert result is None

    def test_get_latest_prerelease(self, mock_repo):
        """Test getting latest prerelease"""
        mock_release = Mock(spec=AbstractRelease)
        mock_repo.get_latest_prerelease.return_value = mock_release

        result = bootstrap.get_latest_prerelease(mock_repo)

        assert result == mock_release
        mock_repo.get_latest_prerelease.assert_called_once()

    def test_find_previous_release_with_prefix(self, mock_repo):
        """Test finding previous release with prefix filter"""
        mock_release1 = Mock(spec=AbstractRelease)
        mock_release1.tag_name = "release/2.0.0"
        mock_release1.prerelease = False

        mock_release2 = Mock(spec=AbstractRelease)
        mock_release2.tag_name = "release/1.0.0"
        mock_release2.prerelease = False

        mock_release3 = Mock(spec=AbstractRelease)
        mock_release3.tag_name = "beta/1.0.0"
        mock_release3.prerelease = True

        mock_repo.releases.return_value = [mock_release1, mock_release2, mock_release3]

        result = bootstrap.find_previous_release(mock_repo, prefix="release/")

        assert result == mock_release2

    def test_find_previous_release_without_prefix(self, mock_repo):
        """Test finding previous release without prefix filter"""
        mock_release1 = Mock(spec=AbstractRelease)
        mock_release1.tag_name = "release/2.0.0"
        mock_release1.prerelease = False

        mock_release2 = Mock(spec=AbstractRelease)
        mock_release2.tag_name = "release/1.0.0"
        mock_release2.prerelease = False

        mock_prerelease = Mock(spec=AbstractRelease)
        mock_prerelease.tag_name = "beta/1.5.0"
        mock_prerelease.prerelease = True

        mock_repo.releases.return_value = [
            mock_release1,
            mock_prerelease,
            mock_release2,
        ]

        result = bootstrap.find_previous_release(mock_repo)

        assert result == mock_release2

    @patch("cumulusci.vcs.bootstrap.get_version_id_from_commit")
    def test_locate_commit_status_package_id_found(
        self, mock_get_version_id, mock_repo
    ):
        """Test locating commit status package ID successfully"""
        mock_branch = Mock(spec=AbstractBranch)
        mock_commit = Mock(spec=AbstractRepoCommit)
        mock_commit.sha = "abc123"
        mock_commit.parents = []
        mock_branch.commit = mock_commit
        mock_get_version_id.return_value = "04t123456789"

        version_id, commit = bootstrap.locate_commit_status_package_id(
            mock_repo, mock_branch, "context_2gp"
        )

        assert version_id == "04t123456789"
        assert commit == mock_commit

    @patch("cumulusci.vcs.bootstrap.get_version_id_from_commit")
    def test_locate_commit_status_package_id_not_found(
        self, mock_get_version_id, mock_repo
    ):
        """Test locating commit status package ID when not found"""
        mock_branch = Mock(spec=AbstractBranch)
        mock_commit = Mock(spec=AbstractRepoCommit)
        mock_commit.sha = "abc123"
        mock_commit.parents = []
        mock_branch.commit = mock_commit
        mock_get_version_id.return_value = None

        version_id, commit = bootstrap.locate_commit_status_package_id(
            mock_repo, mock_branch, "context_2gp"
        )

        assert version_id is None
        assert commit is None  # When no parents, commit becomes None

    @patch("cumulusci.vcs.bootstrap.get_service")
    def test_get_repo_from_config(self, mock_get_service, mock_project_config):
        """Test getting repository from config"""
        mock_service = Mock(spec=VCSService)
        mock_repo = Mock(spec=AbstractRepo)
        mock_service.get_repository.return_value = mock_repo
        mock_get_service.return_value = mock_service

        result = bootstrap.get_repo_from_config(
            mock_project_config, {"option": "value"}
        )

        assert result == mock_repo
        mock_get_service.assert_called_once_with(
            mock_project_config, logger=mock_project_config.logger
        )
        mock_service.get_repository.assert_called_once_with(options={"option": "value"})

    def test_get_latest_tag_production(self, mock_repo):
        """Test getting latest production tag"""
        mock_release = Mock(spec=AbstractRelease)
        mock_release.tag_name = "release/1.0.0"
        mock_repo.latest_release.return_value = mock_release
        mock_repo.project_config = Mock()
        mock_repo.project_config.project__git__prefix_release = "release/"

        result = bootstrap.get_latest_tag(mock_repo, beta=False)

        assert result == "release/1.0.0"

    @patch("cumulusci.vcs.bootstrap._get_latest_tag_for_prefix")
    def test_get_latest_tag_beta(self, mock_get_latest_prefix, mock_repo):
        """Test getting latest beta tag"""
        mock_get_latest_prefix.return_value = "beta/1.0.0"
        mock_repo.project_config = Mock()
        mock_repo.project_config.project__git__prefix_beta = "beta/"

        result = bootstrap.get_latest_tag(mock_repo, beta=True)

        assert result == "beta/1.0.0"
        mock_get_latest_prefix.assert_called_once_with(mock_repo, "beta/")

    def test_get_latest_tag_exception(self, mock_repo):
        """Test getting latest tag when exception occurs"""
        mock_repo.latest_release.side_effect = Exception("Error")
        mock_repo.project_config = Mock()
        mock_repo.project_config.project__git__prefix_release = "release/"
        mock_repo.repo_url = "https://github.com/test/repo"

        with pytest.raises(VcsException) as e:
            bootstrap.get_latest_tag(mock_repo, beta=False)

        assert (
            "No release found for https://github.com/test/repo with tag prefix release/"
            in str(e.value)
        )

    def test_get_latest_tag_for_prefix_success(self, mock_repo):
        """Test getting latest tag for prefix successfully"""
        mock_release1 = Mock(spec=AbstractRelease)
        mock_release1.tag_name = "beta/2.0.0"

        mock_release2 = Mock(spec=AbstractRelease)
        mock_release2.tag_name = "release/1.0.0"

        mock_repo.releases.return_value = [mock_release1, mock_release2]

        result = bootstrap._get_latest_tag_for_prefix(mock_repo, "beta/")

        assert result == "beta/2.0.0"

    def test_get_latest_tag_for_prefix_not_found(self, mock_repo):
        """Test getting latest tag for prefix when not found"""
        mock_release = Mock(spec=AbstractRelease)
        mock_release.tag_name = "release/1.0.0"
        mock_repo.releases.return_value = [mock_release]
        mock_repo.repo_url = "https://github.com/test/repo"

        with pytest.raises(VcsException) as e:
            bootstrap._get_latest_tag_for_prefix(mock_repo, "beta/")

        assert (
            "No release found for https://github.com/test/repo with tag prefix beta/"
            in str(e.value)
        )

    @patch("cumulusci.vcs.bootstrap.get_tag_by_name")
    def test_get_version_id_from_tag_success(self, mock_get_tag, mock_repo):
        """Test getting version ID from tag successfully"""
        mock_tag = Mock(spec=AbstractGitTag)
        mock_tag.message = "Release notes\nversion_id: 04t123456789\nMore info"
        mock_get_tag.return_value = mock_tag

        result = bootstrap.get_version_id_from_tag(mock_repo, "v1.0.0")

        assert result == "04t123456789"
        mock_get_tag.assert_called_once_with(mock_repo, "v1.0.0")

    @patch("cumulusci.vcs.bootstrap.get_tag_by_name")
    def test_get_version_id_from_tag_not_found(self, mock_get_tag, mock_repo):
        """Test getting version ID from tag when not found"""
        mock_tag = Mock(spec=AbstractGitTag)
        mock_tag.message = "Release notes\nNo version id here\nMore info"
        mock_get_tag.return_value = mock_tag

        with pytest.raises(VcsException) as e:
            bootstrap.get_version_id_from_tag(mock_repo, "v1.0.0")

        assert "Could not find version_id for tag v1.0.0" in str(e.value)

    @patch("cumulusci.vcs.bootstrap.get_tag_by_name")
    def test_get_version_id_from_tag_invalid_format(self, mock_get_tag, mock_repo):
        """Test getting version ID from tag with invalid format"""
        mock_tag = Mock(spec=AbstractGitTag)
        mock_tag.message = "Release notes\nversion_id: invalid123\nMore info"
        mock_get_tag.return_value = mock_tag

        with pytest.raises(VcsException) as e:
            bootstrap.get_version_id_from_tag(mock_repo, "v1.0.0")

        assert "Could not find version_id for tag v1.0.0" in str(e.value)


class TestBootstrapIntegration:
    """Integration tests for bootstrap functions"""

    @pytest.fixture()
    def dummy_repo_with_config(self):
        """Create a dummy repo with project config"""
        repo = DummyRepo()
        repo.project_config = Mock()
        repo.project_config.project__git__prefix_release = "release/"
        repo.project_config.project__git__prefix_beta = "beta/"
        repo.repo_url = "https://github.com/test/repo"
        return repo

    def test_get_ref_and_tag_integration(self):
        """Test integration between get_ref_for_tag and get_tag_by_name"""
        repo = DummyRepo()
        tag_name = "v1.0.0"

        # Test the full flow
        ref = bootstrap.get_ref_for_tag(repo, tag_name)
        tag = bootstrap.get_tag_by_name(repo, tag_name)

        assert isinstance(ref, DummyRef)
        assert isinstance(tag, DummyTag)

    @patch("cumulusci.vcs.bootstrap.get_version_id_from_commit")
    def test_locate_commit_status_integration(self, mock_get_version_id):
        """Test integration of commit status location with multiple commits"""
        # Setup commits with parent chain
        commit1 = Mock(spec=AbstractRepoCommit)
        commit1.sha = "abc123"
        commit1.parents = [Mock(sha="def456")]

        commit2 = Mock(spec=AbstractRepoCommit)
        commit2.sha = "def456"
        commit2.parents = []

        mock_branch = Mock(spec=AbstractBranch)
        mock_branch.commit = commit1

        mock_repo = Mock(spec=AbstractRepo)
        mock_repo.get_commit.return_value = commit2

        # First call returns None, second call returns version_id
        mock_get_version_id.side_effect = [None, "04t123456789"]

        version_id, commit = bootstrap.locate_commit_status_package_id(
            mock_repo, mock_branch, "context_2gp"
        )

        assert version_id == "04t123456789"
        assert commit == commit2
        assert mock_get_version_id.call_count == 2


class TestBootstrapRegex:
    """Test regex patterns used in bootstrap"""

    def test_version_id_regex_match(self):
        """Test VERSION_ID_RE regex matching"""
        test_string = "Build status: success\nversion_id: 04t123456789ABC\nMore info"
        match = bootstrap.VERSION_ID_RE.search(test_string)

        assert match is not None
        assert match.group(1) == "04t123456789ABC"

    def test_version_id_regex_no_match(self):
        """Test VERSION_ID_RE regex not matching"""
        test_string = "Build status: success\nNo version id here\nMore info"
        match = bootstrap.VERSION_ID_RE.search(test_string)

        assert match is None


class TestBootstrapEdgeCases:
    """Test edge cases and error conditions"""

    def test_get_pull_requests_empty_list(self):
        """Test getting pull requests when none exist"""
        repo = Mock(spec=AbstractRepo)
        repo.pull_requests.return_value = []

        result = bootstrap.get_pull_requests_with_base_branch(repo, "main")

        assert result == []

    def test_is_label_on_pull_request_empty_labels(self):
        """Test checking label when no labels exist"""
        repo = Mock(spec=AbstractRepo)
        pr = Mock(spec=AbstractPullRequest)
        repo.get_pr_issue_labels.return_value = []

        result = bootstrap.is_label_on_pull_request(repo, pr, "bug")

        assert result is False

    def test_find_previous_release_single_release(self):
        """Test finding previous release when only one exists"""
        repo = Mock(spec=AbstractRepo)
        mock_release = Mock(spec=AbstractRelease)
        mock_release.tag_name = "release/1.0.0"
        mock_release.prerelease = False
        repo.releases.return_value = [mock_release]

        result = bootstrap.find_previous_release(repo)

        assert result is None

    def test_locate_commit_status_no_parents(self):
        """Test locating commit status when commit has no parents"""
        repo = Mock(spec=AbstractRepo)
        branch = Mock(spec=AbstractBranch)
        commit = Mock(spec=AbstractRepoCommit)
        commit.sha = "abc123"
        commit.parents = []
        branch.commit = commit

        with patch(
            "cumulusci.vcs.bootstrap.get_version_id_from_commit", return_value=None
        ):
            version_id, returned_commit = bootstrap.locate_commit_status_package_id(
                repo, branch, "context"
            )

        assert version_id is None
        assert returned_commit is None  # When no parents, commit becomes None



import shutil
from pathlib import Path

from cumulusci.tasks.sfdx import SFDXBaseTask


class DxConvertFrom(SFDXBaseTask):
    task_options = {
        "extra": {"description": "Append additional options to the command"},
        "src_dir": {
            "description": "The path to the src directory where converted contents will be stored. Defaults to src/",
            "required": True,
        },
    }

    def _init_options(self, kwargs):
        super()._init_options(kwargs)

        # append command  -d option to sf} project convert source
        self.options["command"] = f"project convert source -d {self.options['src_dir']}"

    def _run_task(self):
        src_dir = Path(self.options["src_dir"])
        if src_dir.exists():
            shutil.rmtree(src_dir)
        super()._run_task()



from unittest.mock import MagicMock

import pytest
import responses
from simple_salesforce.exceptions import SalesforceMalformedRequest

from cumulusci.tasks.preflight.settings import CheckMyDomainActive, CheckSettingsValue
from cumulusci.tasks.salesforce.tests.util import create_task
from cumulusci.tests.util import CURRENT_SF_API_VERSION

JSON_RESPONSE = {
    "records": [{"IntVal": 3, "FloatVal": 3.0, "BoolVal": True, "StringVal": "foo"}],
    "done": True,
    "totalSize": 1,
}


@responses.activate
@pytest.mark.parametrize(
    "settings_field,value,outcome",
    [
        ("IntVal", 3, True),
        ("FloatVal", 3.0, True),
        ("BoolVal", "true", True),
        ("StringVal", "foo", True),
        ("StringVal", "bad", False),
    ],
)
def test_check_settings(settings_field, value, outcome):
    responses.add(
        "GET",
        f"https://test.salesforce.com/services/data/v{CURRENT_SF_API_VERSION}/tooling/query/?q=SELECT+{settings_field}+FROM+ChatterSettings",
        json=JSON_RESPONSE,
    )
    task = create_task(
        CheckSettingsValue,
        {
            "settings_type": "ChatterSettings",
            "settings_field": settings_field,
            "value": value,
        },
    )

    task()

    assert task.return_values is outcome


@responses.activate
def test_check_settings__no_settings():
    responses.add(
        "GET",
        f"https://test.salesforce.com/services/data/v{CURRENT_SF_API_VERSION}/tooling/query/?q=SELECT+Foo+FROM+ChatterSettings",
        json={"records": []},
    )
    task = create_task(
        CheckSettingsValue,
        {
            "settings_type": "ChatterSettings",
            "settings_field": "Foo",
            "value": True,
        },
    )

    task()

    assert task.return_values is False


@responses.activate
def test_check_settings__failure():
    responses.add(
        "GET",
        status=400,
        url=f"https://test.salesforce.com/services/data/v{CURRENT_SF_API_VERSION}/tooling/query/?q=SELECT+Test+FROM+NoSettings",
        json={},
    )
    task = create_task(
        CheckSettingsValue,
        {
            "settings_type": "NoSettings",
            "settings_field": "Test",
            "value": True,
            "treat_missing_as_failure": True,
        },
    )

    task()

    assert task.return_values is False


@responses.activate
def test_check_settings__exception():
    responses.add(
        "GET",
        status=400,
        url=f"https://test.salesforce.com/services/data/v{CURRENT_SF_API_VERSION}/tooling/query/?q=SELECT+Test+FROM+NoSettings",
        json={},
    )
    task = create_task(
        CheckSettingsValue,
        {
            "settings_type": "NoSettings",
            "settings_field": "Test",
            "value": True,
        },
    )

    with pytest.raises(SalesforceMalformedRequest):
        task()

    assert task.return_values is False


@pytest.mark.parametrize(
    "my_domain,outcome",
    [
        ("https://cumulusci.my.salesforce.com", True),
        ("https://cumulusci.cloudforce.com", True),
        ("https://na44.salesforce.com", False),
    ],
)
def test_my_domain_check(my_domain, outcome):
    org_config = MagicMock()
    org_config.instance_url = my_domain
    task = create_task(CheckMyDomainActive, {}, org_config=org_config)
    task()

    assert task.return_values is outcome



""" Tests for the Command tasks """


from io import BytesIO
from unittest import mock

import pytest
from sarge import Capture

from cumulusci.core.config import (
    BaseProjectConfig,
    OrgConfig,
    TaskConfig,
    UniversalConfig,
)
from cumulusci.core.tests.utils import MockLoggerMixin
from cumulusci.tasks.command import Command, CommandException, SalesforceCommand


class TestCommandTask(MockLoggerMixin):
    """Tests for the basic command task"""

    def setup_method(self):
        self.universal_config = UniversalConfig()
        self.project_config = BaseProjectConfig(
            self.universal_config, config={"noyaml": True}
        )
        self.task_config = TaskConfig()
        self._task_log_handler.reset()
        self.task_log = self._task_log_handler.messages

    def test_functional_run_ls(self):
        """Functional test that actually subprocesses and runs command.

        Checks that command either ran successfully or failed as expected
        so that it can be run on any platform. Other tests will mock
        the subprocess interaction.
        """

        self.task_config.config["options"] = {"command": "ls -la"}

        task = Command(self.project_config, self.task_config)

        # try to run the command and handle the command Exception
        # if no exception, confirm that we logged
        # if exception, confirm exception matches expectations
        try:
            task()
        except CommandException:
            assert any("Return code" in s for s in self.task_log["error"])

        assert any("total" in s for s in self.task_log["info"])

    def test_return_values_success(self):
        """Verify return_values is set when command succeeds"""
        self.task_config.config["options"] = {"command": "ls"}
        task = Command(self.project_config, self.task_config)
        task()
        assert task.return_values == {"returncode": 0}

    def test_return_values_exception(self):
        """Verify return_values is set when command fails"""
        self.task_config.config["options"] = {"command": "exit 1"}
        task = Command(self.project_config, self.task_config)
        with pytest.raises(CommandException):
            task()
        assert task.return_values == {"returncode": 1}

    def test_init__json_env(self):
        task_config = TaskConfig({"options": {"env": "{}", "command": "ls"}})
        task = Command(self.project_config, task_config)
        assert {} == task.options["env"]

    def test_init__dict_env(self):
        task_config = TaskConfig({"options": {"env": {}, "command": "ls"}})
        task = Command(self.project_config, task_config)
        assert {} == task.options["env"]

    def test_get_env__pass_env_false(self):
        task_config = TaskConfig({"options": {"pass_env": "False", "command": "ls"}})
        task = Command(self.project_config, task_config)
        assert {} == task._get_env()

    def test_handle_returncode(self):
        task_config = TaskConfig({"options": {"command": "ls"}})
        task = Command(self.project_config, task_config)
        with pytest.raises(CommandException):
            task._handle_returncode(1, BytesIO(b"err"))

    @mock.patch("cumulusci.tasks.command.sarge")
    def test_run_task(self, sarge):
        p = mock.Mock()
        p.returncode = 0
        p.stdout = Capture()
        p.stdout.add_stream(BytesIO(b"testing testing 123"))
        p.stderr = Capture()
        p.stderr.add_stream(BytesIO(b"e"))
        sarge.Command.return_value = p

        self.task_config.config["options"] = {"command": "ls -la"}
        task = Command(self.project_config, self.task_config)
        task()

        assert any("testing testing" in s for s in self.task_log["info"])


class TestSalesforceCommand:
    def setup_method(self):
        self.universal_config = UniversalConfig()
        self.project_config = BaseProjectConfig(
            self.universal_config, config={"noyaml": True}
        )
        self.task_config = TaskConfig({"options": {"command": "ls"}})
        self.org_config = OrgConfig(
            {"access_token": "TOKEN", "instance_url": "https://na01.salesforce.com"},
            "test",
        )
        self.org_config.refresh_oauth_token = mock.Mock()

    def test_update_credentials(self):
        task = SalesforceCommand(self.project_config, self.task_config, self.org_config)
        task()
        self.org_config.refresh_oauth_token.assert_called_once()

    def test_get_env(self):
        task = SalesforceCommand(self.project_config, self.task_config, self.org_config)
        env = task._get_env()
        assert "SF_ACCESS_TOKEN" in env
        assert "SF_ORG_INSTANCE_URL" in env



import abc
from typing import Any, List, Optional

from cumulusci.core.config.project_config import BaseProjectConfig
from cumulusci.core.dependencies.github import (
    VCS_GITHUB,
    BaseGitHubDependency,
    get_github_repo,
)
from cumulusci.core.dependencies.resolvers import (
    AbstractReleaseTagResolver,
    AbstractTagResolver,
    AbstractUnmanagedHeadResolver,
    AbstractVcsCommitStatusPackageResolver,
    AbstractVcsReleaseBranchResolver,
    DependencyResolutionStrategy,
)
from cumulusci.core.exceptions import DependencyResolutionError
from cumulusci.utils.git import get_feature_branch_name
from cumulusci.vcs.bootstrap import find_repo_feature_prefix
from cumulusci.vcs.github.adapter import GitHubBranch, GitHubRepository


class GitHubTagResolver(AbstractTagResolver):
    """Resolver that identifies a ref by a specific GitHub tag."""

    name = "GitHub Tag Resolver"
    vcs = VCS_GITHUB

    def get_repo(
        self, context: BaseProjectConfig, url: Optional[str]
    ) -> GitHubRepository:
        return get_github_repo(context, url)


class GitHubReleaseTagResolver(AbstractReleaseTagResolver):
    """Resolver that identifies a ref by finding the latest GitHub release."""

    name = "GitHub Release Resolver"
    vcs = VCS_GITHUB

    def get_repo(
        self, context: BaseProjectConfig, url: Optional[str]
    ) -> GitHubRepository:
        return get_github_repo(context, url)


class GitHubBetaReleaseTagResolver(GitHubReleaseTagResolver):
    """Resolver that identifies a ref by finding the latest GitHub release, including betas."""

    name = "GitHub Release Resolver (Betas)"
    include_beta = True


class GitHubUnmanagedHeadResolver(AbstractUnmanagedHeadResolver):
    """Resolver that identifies a ref by finding the latest commit on the main branch."""

    name = "GitHub Unmanaged Resolver"
    vcs = VCS_GITHUB

    def get_repo(
        self, context: BaseProjectConfig, url: Optional[str]
    ) -> GitHubRepository:
        return get_github_repo(context, url)


class GitHubReleaseBranchCommitStatusResolver(AbstractVcsReleaseBranchResolver):
    """Resolver that identifies a ref by finding a beta 2GP package version
    in a commit status on a `feature/NNN` release branch."""

    name = "GitHub Release Branch Commit Status Resolver"
    commit_status_context = "2gp_context"
    commit_status_default = "Build Feature Test Package"
    branch_offset_start = 0
    branch_offset_end = 1

    def get_repo(
        self, context: BaseProjectConfig, url: Optional[str]
    ) -> GitHubRepository:
        return get_github_repo(context, url)


class GitHubReleaseBranchUnlockedResolver(AbstractVcsReleaseBranchResolver):
    """Resolver that identifies a ref by finding an unlocked package version
    in a commit status on a `feature/NNN` release branch."""

    name = "GitHub Release Branch Unlocked Commit Status Resolver"
    commit_status_context = "unlocked_context"
    commit_status_default = "Build Unlocked Test Package"
    branch_offset_start = 0
    branch_offset_end = 1

    def get_repo(
        self, context: BaseProjectConfig, url: Optional[str]
    ) -> GitHubRepository:
        return get_github_repo(context, url)


class GitHubPreviousReleaseBranchCommitStatusResolver(AbstractVcsReleaseBranchResolver):
    """Resolver that identifies a ref by finding a beta 2GP package version
    in a commit status on a `feature/NNN` release branch that is earlier
    than the matching local release branch."""

    name = "GitHub Previous Release Branch Commit Status Resolver"
    commit_status_context = "2gp_context"
    commit_status_default = "Build Feature Test Package"
    branch_offset_start = 1
    branch_offset_end = 3

    def get_repo(
        self, context: BaseProjectConfig, url: Optional[str]
    ) -> GitHubRepository:
        return get_github_repo(context, url)


class GitHubPreviousReleaseBranchUnlockedResolver(AbstractVcsReleaseBranchResolver):
    """Resolver that identifies a ref by finding an unlocked package version
    in a commit status on a `feature/NNN` release branch that is earlier
    than the matching local release branch."""

    name = "GitHub Previous Release Branch Unlocked Commit Status Resolver"
    commit_status_context = "unlocked_context"
    commit_status_default = "Build Unlocked Test Package"
    branch_offset_start = 1
    branch_offset_end = 3

    def get_repo(
        self, context: BaseProjectConfig, url: Optional[str]
    ) -> GitHubRepository:
        return get_github_repo(context, url)


class AbstractGitHubExactMatchCommitStatusResolver(
    AbstractVcsCommitStatusPackageResolver, abc.ABC
):
    """Abstract base class for resolvers that identify a ref by finding a package version
    in a commit status on a branch whose name matches the local branch."""

    def get_repo(
        self, context: BaseProjectConfig, url: Optional[str]
    ) -> GitHubRepository:
        return get_github_repo(context, url)

    def get_branches(
        self,
        dep: BaseGitHubDependency,
        context: BaseProjectConfig,
    ) -> List[GitHubBranch]:
        repo = self.get_repo(context, dep.url)
        if not repo:
            raise DependencyResolutionError(
                f"Unable to access GitHub repository for {dep.url}"
            )

        try:
            remote_branch_prefix = find_repo_feature_prefix(repo)
        except Exception:
            context.logger.info(
                f"Could not find feature branch prefix or commit-status context for {repo.clone_url}. Unable to resolve package."
            )
            return []

        # Attempt exact match
        try:
            branch = get_feature_branch_name(
                context.repo_branch, context.project__git__prefix_feature
            )
            release_branch = repo.branch(f"{remote_branch_prefix}{branch}")
        except Exception:
            context.logger.info(f"Exact-match branch not found for {repo.clone_url}.")
            return []

        return [release_branch]


class GitHubExactMatch2GPResolver(AbstractGitHubExactMatchCommitStatusResolver):
    """Resolver that identifies a ref by finding a 2GP package version
    in a commit status on a branch whose name matches the local branch."""

    name = "GitHub Exact-Match Commit Status Resolver"
    commit_status_context = "2gp_context"
    commit_status_default = "Build Feature Test Package"


class GitHubExactMatchUnlockedCommitStatusResolver(
    AbstractGitHubExactMatchCommitStatusResolver
):
    """Resolver that identifies a ref by finding an unlocked package version
    in a commit status on a branch whose name matches the local branch."""

    name = "GitHub Exact-Match Unlocked Commit Status Resolver"
    commit_status_context = "unlocked_context"
    commit_status_default = "Build Unlocked Test Package"


class AbstractGitHubDefaultBranchCommitStatusResolver(
    AbstractVcsCommitStatusPackageResolver, abc.ABC
):
    """Abstract base class for resolvers that identify a ref by finding a beta package version
    in a commit status on the repo's default branch."""

    def get_repo(
        self, context: BaseProjectConfig, url: Optional[str]
    ) -> GitHubRepository:
        return get_github_repo(context, url)

    def get_branches(
        self,
        dep: BaseGitHubDependency,
        context: BaseProjectConfig,
    ) -> List[GitHubBranch]:
        repo = self.get_repo(context, dep.url)

        return [repo.branch(repo.default_branch)]


class GitHubDefaultBranch2GPResolver(AbstractGitHubDefaultBranchCommitStatusResolver):
    name = "GitHub Default Branch Commit Status Resolver"
    commit_status_context = "2gp_context"
    commit_status_default = "Build Feature Test Package"


class GitHubDefaultBranchUnlockedCommitStatusResolver(
    AbstractGitHubDefaultBranchCommitStatusResolver
):
    name = "GitHub Default Branch Unlocked Commit Status Resolver"
    commit_status_context = "unlocked_context"
    commit_status_default = "Build Unlocked Test Package"


GITHUB_RESOLVER_CLASSES: dict[str, type[Any]] = {
    DependencyResolutionStrategy.STATIC_TAG_REFERENCE: GitHubTagResolver,
    DependencyResolutionStrategy.COMMIT_STATUS_EXACT_BRANCH: GitHubExactMatch2GPResolver,
    DependencyResolutionStrategy.COMMIT_STATUS_RELEASE_BRANCH: GitHubReleaseBranchCommitStatusResolver,
    DependencyResolutionStrategy.COMMIT_STATUS_PREVIOUS_RELEASE_BRANCH: GitHubPreviousReleaseBranchCommitStatusResolver,
    DependencyResolutionStrategy.COMMIT_STATUS_DEFAULT_BRANCH: GitHubDefaultBranch2GPResolver,
    DependencyResolutionStrategy.BETA_RELEASE_TAG: GitHubBetaReleaseTagResolver,
    DependencyResolutionStrategy.RELEASE_TAG: GitHubReleaseTagResolver,
    DependencyResolutionStrategy.UNMANAGED_HEAD: GitHubUnmanagedHeadResolver,
    DependencyResolutionStrategy.UNLOCKED_EXACT_BRANCH: GitHubExactMatchUnlockedCommitStatusResolver,
    DependencyResolutionStrategy.UNLOCKED_RELEASE_BRANCH: GitHubReleaseBranchUnlockedResolver,
    DependencyResolutionStrategy.UNLOCKED_PREVIOUS_RELEASE_BRANCH: GitHubPreviousReleaseBranchUnlockedResolver,
    DependencyResolutionStrategy.UNLOCKED_DEFAULT_BRANCH: GitHubDefaultBranchUnlockedCommitStatusResolver,
}



import pytest

from cumulusci.core.exceptions import TaskOptionsError
from cumulusci.tasks.metadata_etl import SetObjectSettings
from cumulusci.tasks.salesforce.tests.util import create_task
from cumulusci.utils.xml import metadata_tree

MD = "{%s}" % metadata_tree.METADATA_NAMESPACE

CUSTOMOBJECT_SETTINGS_XML = b"""<?xml version="1.0" encoding="UTF-8"?>
<CustomObject xmlns="http://soap.sforce.com/2006/04/metadata">
    <sharingModel>Read</sharingModel>
    <externalSharingModel>Read</externalSharingModel>
    <label>Test</label>
    <pluralLabel>Tests</pluralLabel>
    <nameField>
        <label>Test Name</label>
        <trackHistory>false</trackHistory>
        <type>Text</type>
    </nameField>
    <deploymentStatus>Deployed</deploymentStatus>
    <enableActivities>true</enableActivities>
    <enableBulkApi>true</enableBulkApi>
    <enableFeeds>false</enableFeeds>
    <enableHistory>false</enableHistory>
</CustomObject>"""


class TestSetObjectSettings:
    def test_options__require_enable_or_disable(self):
        with pytest.raises(TaskOptionsError) as e:
            create_task(
                SetObjectSettings,
                {
                    "api_names": "Test__c",
                },
            )
        assert (
            e.value.args[0]
            == "You must provide values for either 'enable' or 'disable'"
        )

    def test_options__invalid_setting(self):
        with pytest.raises(TaskOptionsError) as e:
            create_task(
                SetObjectSettings,
                {
                    "api_names": "Test__c",
                    "enable": ["Foo"],
                },
            )
        assert e.value.args[0].startswith("Invalid settings: Foo.")

    def test_options__invalid_settings(self):
        with pytest.raises(TaskOptionsError) as e:
            create_task(
                SetObjectSettings,
                {
                    "api_names": "Test__c",
                    "enable": ["Foo", "Bar"],
                },
            )
        assert e.value.args[0].startswith("Invalid settings: Foo, Bar.")

    def test_enable__single(self):
        task = create_task(
            SetObjectSettings,
            {
                "api_names": "Test__c",
                "enable": ["Feeds"],
            },
        )
        tree = metadata_tree.fromstring(CUSTOMOBJECT_SETTINGS_XML)

        result = task._transform_entity(tree, "Test__c")

        entry = result._element.findall(f".//{MD}enableFeeds")
        assert len(entry) == 1
        assert entry[0].text == "true"

        entry = result._element.findall(f".//{MD}enableHistory")
        assert len(entry) == 1
        assert entry[0].text == "false"

    def test_enable__single_missing(self):
        task = create_task(
            SetObjectSettings,
            {
                "api_names": "Test__c",
                "enable": ["Sharing"],
            },
        )
        tree = metadata_tree.fromstring(CUSTOMOBJECT_SETTINGS_XML)

        result = task._transform_entity(tree, "Test__c")

        entry = result._element.findall(f".//{MD}enableSharing")
        assert len(entry) == 1
        assert entry[0].text == "true"

    def test_enable__no_change(self):
        task = create_task(
            SetObjectSettings,
            {
                "api_names": "Test__c",
                "enable": ["Activities"],
            },
        )
        tree = metadata_tree.fromstring(CUSTOMOBJECT_SETTINGS_XML)

        result = task._transform_entity(tree, "Test__c")

        entry = result._element.findall(f".//{MD}enableActivities")
        assert len(entry) == 1
        assert entry[0].text == "true"

    def test_enable__multiple(self):
        task = create_task(
            SetObjectSettings,
            {
                "api_names": "Test__c",
                "enable": ["Feeds", "History"],
            },
        )
        tree = metadata_tree.fromstring(CUSTOMOBJECT_SETTINGS_XML)

        result = task._transform_entity(tree, "Test__c")

        entry = result._element.findall(f".//{MD}enableFeeds")
        assert len(entry) == 1
        assert entry[0].text == "true"

        entry = result._element.findall(f".//{MD}enableHistory")
        assert len(entry) == 1
        assert entry[0].text == "true"

    def test_disable__single(self):
        task = create_task(
            SetObjectSettings,
            {
                "api_names": "Test__c",
                "disable": ["Activities"],
            },
        )
        tree = metadata_tree.fromstring(CUSTOMOBJECT_SETTINGS_XML)

        result = task._transform_entity(tree, "Test__c")

        entry = result._element.findall(f".//{MD}enableActivities")
        assert len(entry) == 1
        assert entry[0].text == "false"

        entry = result._element.findall(f".//{MD}enableBulkApi")
        assert len(entry) == 1
        assert entry[0].text == "true"

    def test_disable__single_missing(self):
        task = create_task(
            SetObjectSettings,
            {
                "api_names": "Test__c",
                "disable": ["Sharing"],
            },
        )
        tree = metadata_tree.fromstring(CUSTOMOBJECT_SETTINGS_XML)

        result = task._transform_entity(tree, "Test__c")

        entry = result._element.findall(f".//{MD}enableSharing")
        assert len(entry) == 1
        assert entry[0].text == "false"

    def test_disable__no_change(self):
        task = create_task(
            SetObjectSettings,
            {
                "api_names": "Test__c",
                "disable": ["Feeds"],
            },
        )
        tree = metadata_tree.fromstring(CUSTOMOBJECT_SETTINGS_XML)

        result = task._transform_entity(tree, "Test__c")

        entry = result._element.findall(f".//{MD}enableFeeds")
        assert len(entry) == 1
        assert entry[0].text == "false"

    def test_disable__multiple(self):
        task = create_task(
            SetObjectSettings,
            {
                "api_names": "Test__c",
                "disable": ["Activities", "BulkApi"],
            },
        )
        tree = metadata_tree.fromstring(CUSTOMOBJECT_SETTINGS_XML)

        result = task._transform_entity(tree, "Test__c")

        entry = result._element.findall(f".//{MD}enableActivities")
        assert len(entry) == 1
        assert entry[0].text == "false"

        entry = result._element.findall(f".//{MD}enableBulkApi")
        assert len(entry) == 1
        assert entry[0].text == "false"

    def test_enable_and_disable(self):
        task = create_task(
            SetObjectSettings,
            {
                "api_names": "Test__c",
                "enable": ["Feeds", "History"],
                "disable": ["Activities", "BulkApi"],
            },
        )
        tree = metadata_tree.fromstring(CUSTOMOBJECT_SETTINGS_XML)

        result = task._transform_entity(tree, "Test__c")

        entry = result._element.findall(f".//{MD}enableActivities")
        assert len(entry) == 1
        assert entry[0].text == "false"

        entry = result._element.findall(f".//{MD}enableBulkApi")
        assert len(entry) == 1
        assert entry[0].text == "false"

        entry = result._element.findall(f".//{MD}enableFeeds")
        assert len(entry) == 1
        assert entry[0].text == "true"

        entry = result._element.findall(f".//{MD}enableHistory")
        assert len(entry) == 1
        assert entry[0].text == "true"



import code
import contextlib
import os
import pdb
import runpy
import signal
import sys
import traceback

import click
import requests
import rich
from rich.console import Console
from rich.markup import escape

import cumulusci
from cumulusci.core.debug import set_debug_mode
from cumulusci.core.exceptions import CumulusCIUsageError
from cumulusci.plugins.plugin_loader import load_plugins
from cumulusci.utils import get_cci_upgrade_command
from cumulusci.utils.http.requests_utils import init_requests_trust
from cumulusci.utils.logging import tee_stdout_stderr

from .error import error
from .flow import flow
from .logger import get_tempfile_logger, init_logger
from .org import org
from .plan import plan
from .project import project
from .robot import robot
from .runtime import CliRuntime, pass_runtime
from .service import service
from .task import task
from .utils import (
    check_latest_plugins,
    check_latest_version,
    get_installed_version,
    get_latest_final_version,
    warn_if_no_long_paths,
)

SUGGEST_ERROR_COMMAND = (
    """Run this command for more information about debugging errors: cci error --help"""
)

USAGE_ERRORS = (CumulusCIUsageError, click.UsageError)

# Global variable to track the context stack for cleanup on signal
_exit_stack = None
_signal_handler_active = False  # Flag to prevent recursive signal handler calls


def _cleanup_on_signal(signum):
    """Cleanup action for termination signals."""
    global _signal_handler_active

    # Prevent recursive signal handler calls
    if _signal_handler_active:
        return

    _signal_handler_active = True

    console = Console()

    signal_names = {signal.SIGTERM: "SIGTERM", signal.SIGINT: "SIGINT"}

    signal_name = signal_names.get(signum, f"signal {signum}")

    console.print(
        f"\n[yellow]Received {signal_name} - CumulusCI is being terminated[/yellow]"
    )
    console.print(
        "[yellow]Exiting with failure code due to external cancellation.[/yellow]"
    )

    # Clean up managed resources
    if _exit_stack:
        try:
            _exit_stack.close()
        except Exception as e:
            console.print(f"[red]Error during cleanup: {e}[/red]")

    # Terminate child processes in the process group
    try:
        console.print("[yellow]Terminating child processes...[/yellow]")

        # Temporarily ignore the signal to prevent recursion
        old_handler = signal.signal(signum, signal.SIG_IGN)

        try:
            # Only use process group termination on Unix systems
            if hasattr(os, "getpgrp") and hasattr(os, "killpg"):
                pgrp = os.getpgrp()
                # Send signal to all processes in the group except ourselves
                os.killpg(pgrp, signum)
            else:
                # On Windows, we can't use process groups, so just log the attempt
                console.print(
                    "[yellow]Process group termination not supported on this platform[/yellow]"
                )
        finally:
            # Restore the original signal handler
            signal.signal(signum, old_handler)

    except ProcessLookupError:
        # Process group may not exist or may already be terminated
        pass
    except Exception as e:
        console.print(f"[red]Warning: Error terminating child processes: {e}[/red]")

    # Exit with appropriate failure code
    exit_code = 143 if signum == signal.SIGTERM else 130  # Standard exit codes
    sys.exit(exit_code)


def _signal_handler(signum, frame):
    """Handle termination signals by deferring to the cleanup function."""
    _cleanup_on_signal(signum)


#
# Root command
#
def main(args=None):
    """Main CumulusCI CLI entry point.

    This runs as the first step in processing any CLI command.

    This wraps the `click` library in order to do some initialization and centralized error handling.
    """
    global _exit_stack

    # Create a new process group so we can terminate all child processes
    # when we receive a termination signal
    try:
        if hasattr(os, "setpgrp"):
            # On Unix systems, create a new process group
            os.setpgrp()
    except Exception:
        # On Windows or if setpgrp fails, continue without process group
        pass

    # Set up signal handlers for graceful termination
    signal.signal(signal.SIGTERM, _signal_handler)
    signal.signal(signal.SIGINT, _signal_handler)

    with contextlib.ExitStack() as stack:
        _exit_stack = stack  # Store reference for signal handler cleanup

        args = args or sys.argv

        # (If enabled) set up requests to validate certs using system CA certs instead of certifi
        init_requests_trust()

        # Check for updates _unless_ we've been asked to output JSON,
        # or if we're going to check anyway as part of the `version` command.
        is_version_command = len(args) > 1 and args[1] == "version"
        if "--json" not in args and not is_version_command:
            check_latest_version()
            check_latest_plugins()

        # Only create logfiles for commands that are not `cci error`
        is_error_command = len(args) > 2 and args[1] == "error"
        tempfile_path = None
        if not is_error_command:
            logger, tempfile_path = get_tempfile_logger()
            stack.enter_context(tee_stdout_stderr(args, logger, tempfile_path))

        debug = "--debug" in args
        if debug:
            args.remove("--debug")

        with set_debug_mode(debug):
            try:
                runtime = CliRuntime(load_keychain=False)
            except Exception as e:
                handle_exception(e, is_error_command, tempfile_path, debug)
                sys.exit(1)

            runtime.check_cumulusci_version()
            should_show_stacktraces = runtime.universal_config.cli__show_stacktraces

            init_logger(debug=debug)
            # Hand CLI processing over to click, but handle exceptions
            try:
                cli(args[1:], standalone_mode=False, obj=runtime)
            except click.Abort:  # Keyboard interrupt
                if _exit_stack:
                    _exit_stack.close()
                Console(stderr=True).print("[yellow]Aborted![/yellow]")
                sys.exit(1)
            except Exception as e:
                if debug:
                    console = Console()
                    show_debug_info()
                    console.print(
                        f"\n[red bold]Debug info for bug reports:\n{traceback.format_exc()}"
                    )
                    sys.exit(1)
                else:
                    handle_exception(
                        e, is_error_command, tempfile_path, should_show_stacktraces
                    )
                    sys.exit(1)

        # Clear the global reference when exiting normally
        _exit_stack = None


def handle_exception(
    error,
    is_error_cmd,
    logfile_path,
    should_show_stacktraces=False,
):
    """Displays error of appropriate message back to user, prompts user to investigate further
    with `cci error` commands, and writes the traceback to the latest logfile.
    """
    error_console = Console(stderr=True)
    if isinstance(error, requests.exceptions.ConnectionError):
        connection_error_message(error_console)
    elif isinstance(error, click.ClickException):
        error_console.print(f"[red bold]Error: {escape(error.format_message())}")
    else:
        # We call str ourselves to make Typeguard shut up.
        error_console.print(f"[red bold]Error: {escape(str(error))}")
    # Only suggest gist command if it wasn't run
    if not is_error_cmd:
        error_console.print(f"[yellow]{SUGGEST_ERROR_COMMAND}")

    # This is None if we're handling an exception for a `cci error` command.
    if logfile_path:
        with open(logfile_path, "a") as log_file:
            traceback.print_exc(file=log_file)  # log stacktrace silently

    if should_show_stacktraces and not isinstance(error, USAGE_ERRORS):
        error_console.print_exception()


def connection_error_message(console: Console):
    message = (
        "We encountered an error with your internet connection. "
        "Please check your connection and try the last cci command again."
    )
    console.print(f"[red bold]{message}")


def show_debug_info():
    """Displays the traceback and opens pdb"""
    traceback.print_exc()
    pdb.post_mortem()


def show_version_info():
    console = rich.get_console()
    console.print(
        f"[bold]CumulusCI Plus[/bold] version: {cumulusci.__version__} ({sys.argv[0]})"
    )
    show_plugin_version_infos(console)

    console.print(f"Python version: {sys.version.split()[0]} ({sys.executable})")

    console.print()
    warn_if_no_long_paths(console=console)

    current_version = get_installed_version()
    latest_version = get_latest_final_version()

    if not latest_version > current_version:
        console.print(
            "You have the latest version of CumulusCI Plus :sun_behind_cloud:\n"
        )
        display_release_notes_link(str(latest_version))
        return

    console.print(
        f"[yellow]There is a newer version of CumulusCI Plus available: {str(latest_version)}"
    )
    console.print(f"To upgrade, run `{get_cci_upgrade_command()}`")
    display_release_notes_link(str(latest_version))


def show_plugin_version_infos(console: Console):
    """Display version information for all loaded plugins."""
    plugins = load_plugins()

    if not plugins:
        return

    console.print("Loaded plugins:")

    for plugin in plugins:
        try:
            console.print(f"  [bold]{plugin.name}[/bold]: {plugin.version}")
        except Exception as e:
            console.print(
                f"  [bold]{plugin.name}[/bold]: [red]Error retrieving version: {e}[/red]"
            )


def display_release_notes_link(latest_version: str) -> None:
    """Provide a link to the latest CumulusCI Release Notes"""
    release_notes_link = (
        f"https://github.com/jorgesolebur/CumulusCI/releases/tag/v{latest_version}"
    )
    console = rich.get_console()
    console.print(
        f"See the latest CumulusCI Release Notes: [link={release_notes_link}]{release_notes_link}[/link]"
    )


def version_info_wrapper(
    ctx: click.Context, param: click.Parameter, value: bool
) -> None:
    if not value:
        return
    show_version_info()
    ctx.exit()


@click.group("main", help="")
@click.option(  # based on https://click.palletsprojects.com/en/8.1.x/options/#callbacks-and-eager-options
    "--version",
    is_flag=True,
    expose_value=False,
    is_eager=True,
    help="Show the version and exit.",
    callback=version_info_wrapper,
)
def cli():
    """Top-level `click` command group."""


@cli.command(name="version", help="Print the current version of CumulusCI")
def version():
    show_version_info()


@cli.command(name="shell", help="Drop into a Python shell")
@click.option("--script", help="Path to a script to run", type=click.Path())
@click.option("--python", help="Python code to run directly")
@pass_runtime(require_project=False, require_keychain=True)
def shell(runtime, script=None, python=None):
    # alias for backwards-compatibility
    variables = {
        "config": runtime,
        "runtime": runtime,
        "project_config": runtime.project_config,
    }

    if script:
        if python:
            raise click.UsageError("Cannot specify both --script and --python")
        runpy.run_path(script, init_globals=variables)
    elif python:
        exec(python, variables)
    else:
        code.interact(local=variables)


# Top Level Groups

cli.add_command(error)
cli.add_command(project)
cli.add_command(org)
cli.add_command(service)
cli.add_command(task)
cli.add_command(flow)
cli.add_command(plan)
cli.add_command(robot)



from cumulusci.tasks.salesforce import BaseSalesforceApiTask


class BaseUserLicenseAwareTask(BaseSalesforceApiTask):
    def get_available_user_licenses(self, is_assignable=False):
        """Fetch active user licenses with availability."""
        query = "SELECT Id, LicenseDefinitionKey, TotalLicenses, UsedLicenses FROM UserLicense WHERE Status = 'Active'"
        return {
            lic["Id"]: lic
            for lic in self.sf.query(query)["records"]
            if not is_assignable or (lic["TotalLicenses"] > lic["UsedLicenses"])
        }

    def _log_list(self, title, items):
        self.logger.info(
            f"{title} ({len(items)}):\n" + "\n".join(f"- {item}" for item in items)
        )


class GetAvailableLicenses(BaseUserLicenseAwareTask):
    def _run_task(self):
        self.return_values = [
            result["LicenseDefinitionKey"]
            for result in self.get_available_user_licenses().values()
        ]
        licenses = "\n".join(self.return_values)
        self.logger.info(f"Found licenses:\n{licenses}")


class GetAssignableLicenses(BaseUserLicenseAwareTask):
    def _run_task(self):
        self.return_values = [
            result["LicenseDefinitionKey"]
            for result in self.get_available_user_licenses(is_assignable=True).values()
        ]
        licenses = "\n".join(self.return_values)
        self.logger.info(f"Found assignable licenses:\n{licenses}")


class GetAvailablePermissionSetLicenses(BaseSalesforceApiTask):
    def _run_task(self):
        query = "SELECT PermissionSetLicenseKey FROM PermissionSetLicense WHERE Status = 'Active'"
        self.return_values = [
            result["PermissionSetLicenseKey"]
            for result in self.sf.query(query)["records"]
        ]
        licenses = "\n".join(self.return_values)
        self.logger.info(f"Found permission set licenses:\n{licenses}")


class GetPermissionLicenseSetAssignments(BaseSalesforceApiTask):
    def _run_task(self):
        query = f"SELECT PermissionSetLicense.DeveloperName FROM PermissionSetLicenseAssign WHERE AssigneeId = '{self.org_config.user_id}'"
        self.return_values = [
            result["PermissionSetLicense"]["DeveloperName"]
            for result in self.sf.query_all(query)["records"]
        ]
        permsets = "\n".join(self.return_values)
        self.logger.info(f"Found permission licenses sets assigned:\n{permsets}")


class GetAvailablePermissionSets(BaseSalesforceApiTask):
    def _run_task(self):
        self.return_values = [
            result["Name"]
            for result in self.sf.query_all("SELECT Name FROM PermissionSet")["records"]
        ]
        permsets = "\n".join(self.return_values)
        self.logger.info(f"Found Permission Sets:\n{permsets}")


class GetAssignablePermissionSets(BaseUserLicenseAwareTask):
    def _run_task(self):
        license_data = self.get_available_user_licenses(is_assignable=True)
        permsets = self.sf.query_all("SELECT LicenseId, Name FROM PermissionSet")[
            "records"
        ]
        available_permsets = [
            ps["Name"]
            for ps in permsets
            if not ps["LicenseId"] or ps["LicenseId"] in license_data
        ]

        self.return_values = available_permsets
        permsets = "\n".join(self.return_values)
        self.logger.info(f"Found assignable permission sets:\n{permsets}")



from cumulusci.tasks.bulkdata.extract_dataset_utils.extract_yml import (
    ExtractDeclaration,
)
from cumulusci.tasks.bulkdata.extract_dataset_utils.tests.test_synthesize_extract_declarations import (
    _fake_get_org_schema,
    describe_for,
)
from cumulusci.tasks.bulkdata.generate_mapping_utils.extract_mapping_file_generator import (
    create_extract_mapping_file_from_declarations,
)


class TestGenerateLoadMappingFromDeclarations:
    def test_simple_generate_mapping_from_declarations(self, org_config):
        declarations = [
            ExtractDeclaration(sf_object="Account", fields=["Name", "Description"])
        ]
        object_counts = {"Account": 10, "Contact": 0, "Case": 5}
        obj_describes = (
            describe_for("Account"),
            describe_for("Contact"),
            describe_for("Case"),
        )
        with _fake_get_org_schema(
            org_config,
            obj_describes,
            object_counts,
            filters=[],
            include_counts=True,
        ) as schema:
            mf = create_extract_mapping_file_from_declarations(declarations, schema, ())
            assert mf == {
                "Extract Account": {
                    "api": "smart",
                    "sf_object": "Account",
                    "fields": ["Name", "Description"],
                    "soql_filter": "Name != 'Sample Account for " "Entitlements'",
                }
            }

    def test_generate_mapping_from_declarations__lookups(self, org_config):
        declarations = [
            ExtractDeclaration(sf_object="Account", fields=["Name", "Description"]),
            ExtractDeclaration(sf_object="Contact", fields=["LastName", "AccountId"]),
        ]
        object_counts = {"Account": 10, "Contact": 5, "Case": 5}
        obj_describes = (
            describe_for("Account"),
            describe_for("Contact"),
            describe_for("Case"),
        )
        with _fake_get_org_schema(
            org_config,
            obj_describes,
            object_counts,
            filters=[],
            include_counts=True,
        ) as schema:
            mf = create_extract_mapping_file_from_declarations(declarations, schema, ())
            print(mf)
            assert mf == {
                "Extract Account": {
                    "api": "smart",
                    "sf_object": "Account",
                    "soql_filter": "Name != 'Sample Account for Entitlements'",
                    "fields": ["Name", "Description"],
                },
                "Extract Contact": {
                    "api": "smart",
                    "sf_object": "Contact",
                    "fields": ["LastName"],
                    "lookups": {"AccountId": {"table": ("Account",)}},
                },
            }

    def test_generate_mapping_from_declarations__polymorphic_lookups(self, org_config):
        """Generate correct mapping file for sobjects with polymorphic lookup fields"""
        declarations = [
            ExtractDeclaration(sf_object="Account", fields=["Name", "Description"]),
            ExtractDeclaration(sf_object="Contact", fields=["LastName", "AccountId"]),
            ExtractDeclaration(sf_object="Event", fields=["Subject", "WhoId"]),
            ExtractDeclaration(sf_object="Lead", fields=["LastName", "Company"]),
        ]
        object_counts = {"Account": 10, "Contact": 5, "Case": 5, "Event": 1, "Lead": 2}
        obj_describes = (
            describe_for("Account"),
            describe_for("Contact"),
            describe_for("Case"),
            describe_for("Lead"),
            describe_for("Event"),
        )
        with _fake_get_org_schema(
            org_config,
            obj_describes,
            object_counts,
            filters=[],
            include_counts=True,
        ) as schema:
            mf = create_extract_mapping_file_from_declarations(declarations, schema, ())
            print(mf)
            assert mf == {
                "Extract Account": {
                    "api": "smart",
                    "sf_object": "Account",
                    "soql_filter": "Name != 'Sample Account for Entitlements'",
                    "fields": ["Name", "Description"],
                },
                "Extract Contact": {
                    "api": "smart",
                    "sf_object": "Contact",
                    "fields": ["LastName"],
                    "lookups": {"AccountId": {"table": ("Account",)}},
                },
                "Extract Lead": {
                    "api": "smart",
                    "sf_object": "Lead",
                    "fields": ["LastName", "Company"],
                },
                "Extract Event": {
                    "api": "smart",
                    "sf_object": "Event",
                    "fields": ["Subject"],
                    "lookups": {
                        "WhoId": {
                            "table": (
                                "Contact",
                                "Lead",
                            )
                        }
                    },
                },
            }


# TODO: Figure out where clauses



from cumulusci.core.datasets import Dataset
from cumulusci.core.exceptions import BulkDataException
from cumulusci.tasks.bulkdata.mapping_parser import (
    parse_from_yaml,
    validate_and_inject_mapping,
)
from cumulusci.tasks.bulkdata.step import DataOperationType
from cumulusci.tasks.salesforce import BaseSalesforceApiTask


class LoadDataSetCheck(BaseSalesforceApiTask):
    task_docs = """
        A preflight check to ensure a dataset can be loaded successfully
    """
    task_options = {
        "dataset": {
            "description": "Dataset on which preflight checks need to be performed",
            "required": False,
        },
    }

    def _init_options(self, kwargs):
        super(BaseSalesforceApiTask, self)._init_options(kwargs)
        if "dataset" not in self.options:
            self.options["dataset"] = "default"

    def _run_task(self):
        mapping_file_path = Dataset(
            self.options["dataset"],
            self.project_config,
            self.sf,
            self.org_config,
            schema=None,
        ).mapping_file
        self.mapping = parse_from_yaml(mapping_file_path)
        try:
            validate_and_inject_mapping(
                mapping=self.mapping,
                sf=self.sf,
                namespace=self.project_config.project__package__namespace,
                data_operation=DataOperationType.INSERT,
                inject_namespaces=True,
                drop_missing=False,
            )
            self.return_values = True
        except BulkDataException as e:
            self.logger.error(e)
            self.return_values = False
        return self.return_values



import logging
from contextlib import contextmanager
from unittest import mock

import pytest

from cumulusci.core.datasets import Dataset as RealDataset
from cumulusci.core.exceptions import TaskOptionsError
from cumulusci.tasks.bulkdata.extract_dataset_utils.tests.test_synthesize_extract_declarations import (
    _fake_get_org_schema,
    describe_for,
)
from cumulusci.tasks.sample_data.load_sample_data import LoadSampleData


@contextmanager
def setup_test(org_config):
    object_counts = {"Account": 6, "Contact": 1, "Opportunity": 5}
    obj_describes = (
        describe_for("Account"),
        describe_for("Contact"),
        describe_for("Opportunity"),
    )
    with mock.patch.object(
        type(org_config), "is_person_accounts_enabled", False
    ), mock.patch(
        "cumulusci.tasks.sample_data.load_sample_data.get_org_schema",
        lambda _sf, org_config, **kwargs: _fake_get_org_schema(
            org_config,
            obj_describes,
            object_counts,
            included_objects=["Account", "Contact", "Opportunity"],
            **kwargs,
        ),
    ):
        yield


class TestLoadDatasets:
    @mock.patch("cumulusci.tasks.sample_data.load_sample_data.Dataset")
    def test_simple_extract(
        self,
        Dataset,
        create_task,
        org_config,
    ):
        with setup_test(org_config):
            org_config.config["config_name"] = "dev"
            task = create_task(LoadSampleData, {"ignore_row_errors": True})
            task()
            # default dataset should be opened
            Dataset.assert_any_call(
                org_config.config_name, mock.ANY, mock.ANY, org_config, mock.ANY
            )
            # and loaded
            Dataset().__enter__().load.assert_called_with(
                {"ignore_row_errors": True}, task.logger
            )

    @mock.patch("cumulusci.tasks.sample_data.load_sample_data.Dataset")
    def test_simple_extract__name(
        self,
        Dataset,
        create_task,
        org_config,
    ):
        with setup_test(org_config):
            task = create_task(LoadSampleData, {"dataset": "mydataset"})
            task()
            # default dataset should opened
            Dataset.assert_any_call(
                "mydataset", mock.ANY, mock.ANY, org_config, mock.ANY
            )
            # and loaded
            Dataset().__enter__().load.assert_called_with(
                {"dataset": "mydataset"}, task.logger
            )

    @mock.patch("cumulusci.tasks.sample_data.load_sample_data.Dataset")
    def test_simple_extract__scratch_config(
        self, Dataset, sf, create_task, org_config, project_config
    ):
        ds = RealDataset(
            "qwerty",
            project_config,
            sf,
            org_config,
        )
        with setup_test(org_config):
            org_config.config["config_name"] = "qwerty"
            ds.path.mkdir()
            try:
                task = create_task(LoadSampleData, {})
                task()
                # default dataset should opened
                Dataset.assert_any_call(
                    "qwerty", mock.ANY, mock.ANY, org_config, mock.ANY
                )
                # and loaded
                Dataset().__enter__().load.assert_called_with({}, task.logger)
            finally:
                ds.delete()

    def test_no_match_to_name(self, create_task):
        task = create_task(LoadSampleData, {"dataset": "f.z.f.efa.f"})
        with pytest.raises(TaskOptionsError, match="Could not find.*f.z.f.efa.f"):
            task()

    @mock.patch(
        "cumulusci.tasks.sample_data.load_sample_data.Dataset", spec=RealDataset
    )
    def test_no_dataset_available(self, Dataset, caplog, create_task, org_config):
        caplog.set_level(logging.INFO)
        with setup_test(org_config):
            org_config.config["config_name"] = "qa"
            task = create_task(LoadSampleData, {})
            Dataset().exists.return_value = False
            assert not Dataset().exists()
            task()
            Dataset().__enter__.assert_not_called()
            assert "No contextual sample data found" in caplog.text
